# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
from typing import Union


def replace_enum_values(dd: {}) -> None:
    """Replace enum values with their text strings."""
    to_update = {key: val.name for key, val in dd.items() if isinstance(val, enum.Enum)}
    if to_update:
        dd.update(to_update)


def prop_to_enum(props: {}, enum_types: ()) -> None:
    for p_name, p_type in enum_types:
        if isinstance(props[p_name], str):
            props[p_name] = p_type[props[p_name]]


__probe_value_hex_to_bin__ = {
    "0": "0000",
    "1": "0001",
    "2": "0010",
    "3": "0011",
    "4": "0100",
    "5": "0101",
    "6": "0110",
    "7": "0111",
    "8": "1000",
    "9": "1001",
    "a": "1010",
    "b": "1011",
    "c": "1100",
    "d": "1101",
    "e": "1110",
    "f": "1111",
    "x": "xxxx",
    "_": "",  # Don't keep underscore.
}

# 4-bit values created with { f'{idx:04b}'[::-1]: hex(idx)[2] for idx in range(16) }
__bin_reversed_to_hex__ = {
    # Keys: binary values with reversed bit order.
    "0": "0",
    "1": "1",
    "00": "0",
    "10": "1",
    "01": "2",
    "11": "3",
    "000": "0",
    "100": "1",
    "010": "2",
    "110": "3",
    "001": "4",
    "101": "5",
    "011": "6",
    "111": "7",
    "0000": "0",
    "1000": "1",
    "0100": "2",
    "1100": "3",
    "0010": "4",
    "1010": "5",
    "0110": "6",
    "1110": "7",
    "0001": "8",
    "1001": "9",
    "0101": "A",
    "1101": "B",
    "0011": "C",
    "1011": "D",
    "0111": "E",
    "1111": "F",
}


__hex_char_reversed__ = {
    "0": "0",
    "1": "8",
    "2": "4",
    "3": "C",
    "4": "2",
    "5": "A",
    "6": "6",
    "7": "E",
    "8": "1",
    "9": "9",
    "A": "5",
    "B": "D",
    "C": "3",
    "D": "B",
    "E": "7",
    "F": "F",
}


def bin_reversed_to_hex(bin_value: str) -> str:
    hex_chars = [
        __bin_reversed_to_hex__[bin_value[idx : idx + 4]]
        for idx in reversed(range(0, len(bin_value), 4))
    ]
    return "".join(hex_chars)


def int_reversed_to_hex(val: int, bit_width: int) -> str:
    # 4 bit_len examples:  4 => '2', 3 -> 'C'
    # 3 bit_len examples:  4 => '1', 3 -> '6'
    bin_value = f"{val:0{bit_width}b}"
    return bin_reversed_to_hex(bin_value)


def bin_reversed_to_hex_values(bin_values: [str]) -> [str]:
    return [bin_reversed_to_hex(val) for val in bin_values]


def to_bin_str(val: Union[int, str], bit_width: int, enum_def: enum.EnumMeta = None) -> str:

    if isinstance(val, enum.Enum):
        val = val.value
    elif enum_def and type(val) == str and not (val.startswith("0x") or val.startswith("0X")):
        try:
            val = enum_def[val].value
        except ValueError:
            # Keep going. val is not a enum name.
            pass
    if isinstance(val, int):
        # Both positive and negative integers, with '0'/'1' fill.
        return f"{val:0{bit_width}b}" if val >= 0 else bin((1 << bit_width) + val)[2:]
    if val.startswith("0x") or val.startswith("0X"):
        bin_val = "".join([__probe_value_hex_to_bin__[ch] for ch in val[2:].lower()])
        start_index = len(bin_val) - bit_width
        return bin_val[start_index:]
    else:
        # Already a binary string.
        return val


def bytearray_bit_reversed_to_hex(buffer: bytearray) -> str:
    """Bit-order is reversed for buffer and converted to hex-string."""
    # Hex string in reverse byte order. LSB will be first in 'hex_list'.
    hex_list = list(buffer.hex().upper())
    # Swap upper/lower hex char, within each byte.
    for idx in range(0, len(hex_list), 2):
        hex_list[idx], hex_list[idx + 1] = hex_list[idx + 1], hex_list[idx]
    # Reverse bits in each hex char.
    hex_reversed = "".join([__hex_char_reversed__[ch] for ch in hex_list])
    return hex_reversed


__byte_bit_reversed__ = None


def _get_reversed_byte(bb: int):
    global __byte_bit_reversed__
    if not __byte_bit_reversed__:
        __byte_bit_reversed__ = [int(f"{bb:08b}"[::-1], 2) for bb in range(256)]
    return __byte_bit_reversed__[bb]


def bytearray_each_32bit_word_reversed_to_hex(buffer: bytearray) -> str:
    """Bit-order for each 32 bits are reversed and converted to hex-string."""
    if len(buffer) % 4 != 0:
        raise IndexError("Data is not aligned on 4 byte boundary.")
    buf2 = bytearray(len(buffer))
    for idx in range(0, len(buffer), 4):
        buf2[idx] = _get_reversed_byte(buffer[idx + 3])
        buf2[idx + 1] = _get_reversed_byte(buffer[idx + 2])
        buf2[idx + 2] = _get_reversed_byte(buffer[idx + 1])
        buf2[idx + 3] = _get_reversed_byte(buffer[idx])

    hex_list = [f"{bb:02X}" for bb in reversed(buf2)]
    return "".join(hex_list)


def round_up_to_power_of_two(num: int) -> int:
    return 1 if num == 0 else 1 << (num - 1).bit_length()


def round_down_to_power_of_two(num: int) -> int:
    return 1 if num == 0 else 1 << (num.bit_length() - 1)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
from datetime import datetime
from time import sleep
from abc import ABC, abstractmethod

from chipscopy.dm import NodeListener
from chipscopy.dm.harden.noc_perfmon.noc_types import (
    noc_nmu_typedef,
    noc_nsu_typedef,
    ddrmc_noc_typedef,
    noc_node_types,
    ddrmc_main_typedef,
)
from chipscopy.utils.logger import log

# %%
DOMAIN = "noc_perfmon"
epoch = datetime.utcfromtimestamp(0)  # this is stupid, be best python

# %%
# defines
MAX_SAMPLES = 50
ch_map = {"r": 0, "w": 1}


# %%
def get_noc_typedef_from_name(name) -> str:
    nmu_re = re.compile(".*NMU.*", flags=re.IGNORECASE)
    nsu_re = re.compile(".*NSU.*", flags=re.IGNORECASE)
    ddrmc_noc_re = re.compile(".*DDRMC_NOC.*", flags=re.IGNORECASE)
    ddrmc_main_re = re.compile(".*DDRMC_MAIN.*", flags=re.IGNORECASE)
    ddrmc_re = re.compile(".*DDRMC_X.*", flags=re.IGNORECASE)

    if nmu_re.match(name):
        return noc_nmu_typedef
    elif nsu_re.match(name):
        return noc_nsu_typedef
    elif ddrmc_noc_re.match(name):
        return ddrmc_noc_typedef
    elif ddrmc_main_re.match(name) or ddrmc_re.match(name):
        return ddrmc_main_typedef
    else:
        # log[DOMAIN].error(f'unknown type for name: {name}')
        return ""


# %%
class NoCElement(ABC):
    """
    Goal of this class is to serve as storage for the data in nodeUpdate events
    """

    def __init__(
        self,
        name,
        alt_name,
        num_samples,
        sampling_period_ms,
        record_to_file,
        tslide=0,
        counter_freq_mhz=1000.0,
    ):
        self.name = name
        self.alt_name = alt_name
        self.num_samples = num_samples
        self.sampling_period_ms = sampling_period_ms
        self.record_to_file = record_to_file
        self.node_type = get_noc_typedef_from_name(self.name)
        self.tslide = tslide
        self.counter_freq_mhz = counter_freq_mhz

        self.samples = {
            "read_bandwidth": [0] * MAX_SAMPLES,
            "write_bandwidth": [0] * MAX_SAMPLES,
            "avg_read_latency": [0] * MAX_SAMPLES,
            "avg_write_latency": [0] * MAX_SAMPLES,
            "flags": [],
            "ts": [],
            "type": self.node_type,
        }
        self.raw_trace_data = {}

    @abstractmethod
    def update_node(self, tcf_node, updated_keys):  # pragma: no cover
        pass

    def record_bw_lat(
        self,
        read_bytes,
        write_bytes,
        read_bytes_per_s,
        write_bytes_per_s,
        avg_read_latency,
        avg_write_latency,
    ):
        self.samples["read_bandwidth"].append(read_bytes_per_s)
        self.samples["write_bandwidth"].append(write_bytes_per_s)
        self.samples["avg_read_latency"].append(avg_read_latency)
        self.samples["avg_write_latency"].append(avg_write_latency)
        log[DOMAIN].info(
            f"{self.name}: rb: {read_bytes}, wb: {write_bytes}, "
            + f"rbw: {read_bytes_per_s}, wbw: {write_bytes_per_s}, "
            + f"arl: {avg_read_latency}, awl: {avg_write_latency}"
        )

    def trim_and_log(self, raw_trace_data):
        # trim samples down:
        for _, sample_data in self.samples.items():
            if len(sample_data) > MAX_SAMPLES:
                sample_data.pop(0)

        if self.record_to_file:
            date_handler = lambda obj: (
                obj.isoformat()
                if isinstance(obj, (datetime, datetime.date))
                else json.JSONEncoder().default(obj)
            )
            with open("raw_trace_data.json", "a") as ofh:
                json.dump(raw_trace_data, ofh, default=date_handler)
                ofh.write("\n")
            with open("scaled_trace_data.json", "a") as ofh:
                data_to_dump = {self.name: self.samples}
                json.dump(data_to_dump, ofh, default=date_handler)
                ofh.write("\n")

        self.raw_trace_data = raw_trace_data


# %%
class DDRMC(NoCElement):
    dc_metrics_map = {
        0: "activates",
        1: "read_cas",
        2: "write_cas",
        3: "precharge",
        4: "precharge_all",
        5: "refresh",
        6: "idle",
        7: "overhead",
        8: "bus_turnarounds",
    }

    def __init__(self, *args):
        """
        for DDRMC 'ddrmc_main_X' shall be the name and 'ddrmc_noc_X' shall serve as the alternate name
        :param name:
        :param alt_name:
        """
        super().__init__(*args)
        # add DC aggregate, NSU agg, and NSU port metrics
        for ch in range(0, 2):
            for metric in DDRMC.dc_metrics_map.values():
                self.samples.update({f"dc{ch}_{metric}": [0] * MAX_SAMPLES})
                if ch == 0:
                    self.samples.update({f"agg_{metric}": [0] * MAX_SAMPLES})
            self.samples.update({f"dc{ch}_flags": [0] * MAX_SAMPLES})

        for nsu in range(0, 4):
            self.samples.update({f"nsu{nsu}_write_bandwidth": [0] * MAX_SAMPLES})
            self.samples.update({f"nsu{nsu}_read_bandwidth": [0] * MAX_SAMPLES})
            self.samples.update({f"nsu{nsu}_write_latency": [0] * MAX_SAMPLES})
            self.samples.update({f"nsu{nsu}_read_latency": [0] * MAX_SAMPLES})

        self.axis_metrics = {
            "left_axis": {
                "bandwidth": ["write_bandwidth", "read_bandwidth"],
                "latency": ["avg_write_latency", "avg_read_latency"],
                "ddrmc": [
                    "agg_activates",
                    "agg_read_cas",
                    "agg_write_cas",
                    "agg_precharge",
                    "agg_precharge_all",
                    "agg_refresh",
                    "agg_idle",
                    "agg_overhead",
                    "agg_bus_turnarounds",
                ],
            },
            "right_axis": {"bandwidth": [], "latency": [], "ddrmc": []},
        }

    def update_node(self, tcf_node, updated_keys):
        self.num_samples -= 1
        raw_trace_data = {
            "type": self.node_type,
            "name": self.name,
            # 'ts': (datetime.now() - self.start_time) + epoch,
            "ts": datetime.now(),
            "r": {},
            "w": {},
        }

        tcf_node_type = get_noc_typedef_from_name(tcf_node.Name)

        if tcf_node_type == ddrmc_noc_typedef:
            for ch, mon in ch_map.items():
                raw_trace_data[ch]["lacc"] = []
                raw_trace_data[ch]["burst_count"] = []
                raw_trace_data[ch]["byte_count"] = []
                raw_trace_data[ch]["flags"] = []
                for nsu in range(0, 4):
                    lacc = tcf_node[f"nsu{nsu}_perf_mon_{mon}_0"]
                    raw_trace_data[ch]["lacc"].append(lacc & 0x7FFFFFFF)
                    raw_trace_data[ch]["burst_count"].append(tcf_node[f"nsu{nsu}_perf_mon_{mon}_1"])
                    raw_trace_data[ch]["byte_count"].append(tcf_node[f"nsu{nsu}_perf_mon_{mon}_2"])
                    raw_trace_data[ch]["flags"].append((lacc & 0x80000000) >> 31)

            # bandwidth
            rf = 0
            wf = 0
            for nsu in range(0, 4):
                rf += raw_trace_data["r"]["burst_count"][nsu]
                wf += raw_trace_data["w"]["burst_count"][nsu]
            read_bytes = rf * 128 / 8
            write_bytes = wf * 128 / 8
            read_bytes_per_s = read_bytes / (self.sampling_period_ms / 1000)
            write_bytes_per_s = write_bytes / (self.sampling_period_ms / 1000)

            # flags
            flags = []
            for ch in ["r", "w"]:
                flag = 0
                for nsu in range(0, 4):
                    flag += raw_trace_data[ch]["flags"][nsu]
                if flag != 0:
                    flags.append(f"{ch}_overflow")
            self.samples["flags"].append(flags)

            # latency
            avg_lat = {"r": 0, "w": 0}
            for ch in ["r", "w"]:
                agg_lat = 0
                agg_burst = 0
                for nsu in range(0, 4):
                    agg_lat += raw_trace_data[ch]["lacc"][nsu]
                    agg_burst += raw_trace_data[ch]["burst_count"][nsu]
                if agg_burst == 0:
                    agg_burst = 1
                avg_lat[ch] = float(agg_lat) / float(agg_burst)

            # common op
            self.record_bw_lat(
                read_bytes,
                write_bytes,
                read_bytes_per_s,
                write_bytes_per_s,
                avg_lat["r"],
                avg_lat["w"],
            )

        if tcf_node_type == ddrmc_main_typedef:
            for ch in range(0, 2):
                flags = 0
                for m_index, metric in DDRMC.dc_metrics_map.items():
                    agg_metric = f"agg_{metric}"
                    disagg_metric = f"dc{ch}_{metric}"
                    self.samples[disagg_metric].append(0)
                    if ch == 0:
                        self.samples[agg_metric].append(0)
                    data = tcf_node[f"dc{ch}_perf_mon_{m_index}"]
                    raw_trace_data[disagg_metric] = data
                    overflow = data & 0x80000000 == 0x80000000
                    data &= 0x7FFF_FFFF
                    if overflow:
                        flags |= 1 << m_index  # overflow
                    self.samples[disagg_metric][-1] = data
                    self.samples[agg_metric][-1] += data
                    # need to trim here or we can get a dim mismatch when plotting happens
                    # unique issue for ddmrc, since it's two nodes with polls coming in at different rates
                    if len(self.samples[disagg_metric]) > MAX_SAMPLES:
                        self.samples[disagg_metric].pop(0)
                    if ch == 1:
                        if len(self.samples[agg_metric]) > MAX_SAMPLES:
                            self.samples[agg_metric].pop(0)
                self.samples[f"dc{ch}_flags"].append(flags)

        # timestamp handling
        # the way the polls are created the NA reports AFTER the main, so when this event is received, update plots
        if tcf_node_type == ddrmc_noc_typedef:
            time_delta = raw_trace_data["ts"].strftime("%M:%S.") + str(
                int(raw_trace_data["ts"].microsecond / 1000)
            )
            self.samples["ts"].append(time_delta)
            self.trim_and_log(raw_trace_data)


# %%
class NoCMasterSlave(NoCElement):
    def __init__(self, *args):
        super().__init__(*args)
        self.samples.update(
            {
                "min_r_latency": [0] * MAX_SAMPLES,
                "max_r_latency": [0] * MAX_SAMPLES,
                "min_w_latency": [0] * MAX_SAMPLES,
                "max_w_latency": [0] * MAX_SAMPLES,
            }
        )
        self.axis_metrics = {
            "left_axis": {
                "bandwidth": ["write_bandwidth", "read_bandwidth"],
                "latency": [
                    "avg_write_latency",
                    "avg_read_latency",
                    "min_w_latency",
                    "max_w_latency",
                    "min_r_latency",
                    "max_r_latency",
                ],
                "ddrmc": [],
            },
            "right_axis": {"bandwidth": [], "latency": [], "ddrmc": []},
        }

    def update_node(self, tcf_node, updated_keys):
        raw_trace_data = {
            "type": self.node_type,
            "name": self.name,
            # 'ts': (datetime.now() - self.start_time) + epoch,
            "ts": datetime.now(),
            "r": {},
            "w": {},
        }
        for ch, mon in ch_map.items():
            raw_trace_data[ch]["burst_count"] = tcf_node[f"reg_perf_mon{mon}_burst_cnt"]
            raw_trace_data[ch]["byte_count"] = tcf_node[f"reg_perf_mon{mon}_byte_cnt_lwr"] + (
                (tcf_node[f"reg_perf_mon{mon}_cnt_and_ofl"] & 0xFF) << 32
            )
            raw_trace_data[ch]["flags"] = (tcf_node[f"reg_perf_mon{mon}_cnt_and_ofl"] & 0x1F00) >> 8
            raw_trace_data[ch]["lacc"] = (
                tcf_node[f"reg_perf_mon{mon}_latency_acc_lwr"]
                + ((tcf_node[f"reg_perf_mon{mon}_latency_acc_upr"] & 0xFF) << 32)
            ) << self.tslide
            raw_trace_data[ch]["lmax"] = tcf_node[f"reg_perf_mon{mon}_latency_max"] << self.tslide
            raw_trace_data[ch]["lmin"] = tcf_node[f"reg_perf_mon{mon}_latency_min"] << self.tslide

        # bandwidth
        read_bytes = raw_trace_data["r"]["byte_count"]
        write_bytes = raw_trace_data["w"]["byte_count"]
        read_bytes_per_s = read_bytes * 1000 / self.sampling_period_ms
        write_bytes_per_s = write_bytes * 1000 / self.sampling_period_ms
        # latency and flags
        flags = []
        for ch in ["r", "w"]:
            self.samples[f"min_{ch}_latency"].append(raw_trace_data[ch]["lmin"])
            self.samples[f"max_{ch}_latency"].append(raw_trace_data[ch]["lmax"])
            # flags
            raw_flags = raw_trace_data[ch]["flags"]
            if raw_flags != 0:
                if (raw_flags & 0x10) == 0x10:
                    flags.append(f"{ch}_byte_count_overflow")
                if (raw_flags & 0x08) == 0x08:
                    flags.append(f"{ch}_burst_count_overflow")
                if (raw_flags & 0x04) == 0x04:
                    flags.append(f"{ch}_accumulated_latency_overflow")
                if (raw_flags & 0x02) == 0x02:
                    flags.append(f"{ch}_max_latency_overflow")
                if (raw_flags & 0x01) == 0x01:
                    flags.append(f"{ch}_min_latency_overflow")
        self.samples["flags"].append(flags)

        # latency
        avg_lat = {"r": 0, "w": 0}
        for ch in ["r", "w"]:
            avg_lat[ch] = raw_trace_data[ch]["lacc"] / (
                raw_trace_data[ch]["burst_count"] if raw_trace_data[ch]["burst_count"] != 0 else 1
            )
        self.record_bw_lat(
            read_bytes, write_bytes, read_bytes_per_s, write_bytes_per_s, avg_lat["r"], avg_lat["w"]
        )
        time_delta = raw_trace_data["ts"].strftime("%M:%S.") + str(
            int(raw_trace_data["ts"].microsecond / 1000)
        )
        self.samples["ts"].append(time_delta)
        self.trim_and_log(raw_trace_data)


class NoCPerfMonNodeListener(NodeListener):
    """
    Example Listener class that will print out metrics from NoC nodes (NMU|NSU|DDRMC)

    """

    def __init__(
        self,
        sampling_period_ms,
        num_samples,
        enabled_nodes,
        record_to_file,
        npi_clk_freq=299997009,
        noc_clk_freq=999989990,
        extended_monitor_config=None,
    ):
        self.sampling_period_ms = sampling_period_ms
        self.num_samples = {}
        self.npi_clk_freq = npi_clk_freq
        self.noc_clk_freq = noc_clk_freq
        self.plotter = None
        super().__init__()
        # setup logging
        log.enable_domain([DOMAIN])
        log.change_log_level("WARNING")
        # self.start_time = datetime.now()
        self.noc_elements = {}  # storage for the active monitors
        self.unique_elements = (
            {}
        )  # same objects from above, but with alt mapping removed -- for plot utils

        # build data structure for holding historical data
        for elem in enabled_nodes:
            tslide = 0
            if extended_monitor_config is not None:
                if elem in extended_monitor_config.keys():
                    node_config = extended_monitor_config[elem]
                    if "tslide" in node_config:
                        tslide = node_config["tslide"]
            node_type = get_noc_typedef_from_name(elem)
            if node_type in [noc_nmu_typedef, noc_nsu_typedef]:
                storage_node = NoCMasterSlave(
                    elem.lower(),
                    None,
                    num_samples,
                    sampling_period_ms["NPI"],
                    record_to_file,
                    tslide,
                    npi_clk_freq / 1000000.0,
                )
                self.noc_elements[elem.lower()] = storage_node
                self.unique_elements[elem.lower()] = storage_node
            elif node_type in [ddrmc_main_typedef]:
                alt_name = "ddrmc_noc_" + elem.split("_")[-1]
                node = DDRMC(
                    elem.lower(),
                    alt_name.lower(),
                    num_samples,
                    sampling_period_ms["NoC"],
                    record_to_file,
                    tslide,
                    noc_clk_freq / 1000000.0,
                )
                self.noc_elements[elem.lower()] = node
                self.noc_elements[alt_name.lower()] = node
                self.unique_elements[elem.lower()] = node
            else:  # pragma: no cover
                raise Exception(f"Error, unsupported node: {elem}, type: {node_type}")

    def node_changed(self, node, updated_keys):
        node_type = get_noc_typedef_from_name(node.Name)
        name = node.Name.lower()
        if node_type in noc_node_types and node.Name.lower() in self.noc_elements.keys():
            log[DOMAIN].debug(f"{node.type}: {node.Name}")

            # dispatch the class handler
            self.noc_elements[name].update_node(node, updated_keys)

            # finally update plots
            if self.plotter is not None and node_type is not ddrmc_main_typedef:
                self.plotter.update(self.noc_elements[name])

    def link_plotter(self, plotter):
        self.plotter = plotter
        self.plotter.listener = self


class PerfTGController:
    def __init__(self, base_address, device, vio=None):
        self.base_address = base_address
        self.device = device
        self.vio = vio
        self.instr_ba = self.base_address + 0x8000
        self.ctrl = self.base_address + 0x4000
        self.tg_start = self.base_address + 0x4004
        self.instr_0 = []
        self.instr_1 = []
        self.supported_patterns = {
            "Write Linear": [
                [
                    0xFF200000,
                    0x00080000,
                    0x00000002,
                    0x00000000,
                    0xFFE00000,
                    0x000FFFFF,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00010020,
                    0x00143500,
                    0x0000048C,
                    0x0,
                ],
                [
                    0x00000001,
                    0x00100000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00020000,
                    0x00000000,
                    0x00000000,
                    0x0,
                ],
            ],
            "Read Linear": [
                [
                    0xFF200000,
                    0x00000000,
                    0x00000002,
                    0x00000000,
                    0xFFE00000,
                    0x000FFFFF,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00010020,
                    0x00143500,
                    0x0000048E,
                    0x0,
                ],
                [
                    0x00000001,
                    0x00100000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00000000,
                    0x00020000,
                    0x00000000,
                    0x00000000,
                    0x0,
                ],
            ],
        }
        self.active_pattern = "Write Linear"
        self.trigger_block_rst_probe = None
        self.tg_rst_probe = None
        for probe in self.vio.probe_names:
            if probe.split("/")[-1] == "noc_sim_trig_rst_n":
                self.trigger_block_rst_probe = probe
            if probe.split("/")[-1] == "noc_tg_tg_rst_n":
                self.tg_rst_probe = probe

    def start(self):
        self.device.memory_write(self.tg_start, [0x1])
        # CR-1062874
        # Second write of the ctrl reg is required for the start command to be effective.
        sleep(0.5)
        self.device.memory_write(self.tg_start, [0x1])

    def stop(self):
        # this doesn't seem to be doing much
        self.device.memory_write(self.tg_start, [0x0])

        self._soft_reset()
        self._clear_instruction_memory()

    def block_reset(self):

        if self.vio is not None:
            self.vio.write_probes(
                {
                    self.trigger_block_rst_probe: 0x0,
                    self.tg_rst_probe: 0x0,
                }
            )
            self.vio.write_probes(
                {
                    self.trigger_block_rst_probe: 0x1,
                    self.tg_rst_probe: 0x1,
                }
            )

    def program(self):
        self._soft_reset()
        self.device.memory_write(self.instr_ba, self.supported_patterns[self.active_pattern][0])
        self.device.memory_write(
            self.instr_ba + 0x40, self.supported_patterns[self.active_pattern][1]
        )

    def set_active_pattern(self, pattern):
        if pattern not in self.supported_patterns.keys():
            raise ValueError(f"Unsupported pattern: {pattern}")
        else:
            self.active_pattern = pattern

    def _clear_instruction_memory(self):
        self.device.memory_write(self.instr_ba, [0x0] * 13)
        self.device.memory_write(self.instr_ba + 0x40, [0x0] * 13)

    def _soft_reset(self):
        self.device.memory_write(self.ctrl, [0x0])
        self.device.memory_write(self.ctrl, [0x1])



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from chipscopy.tcf import services
from chipscopy.proxies.CorePropertyProxy import CorePropertyProxy
from chipscopy.utils.logger import log

# NOTE: This name MUST match with the services name defined in the server side AXIS VIO code
AXIS_VIO_SERVICE_NAME = "AxisVIO"
AXIS_VIO_NODE_NAME = "vio"

# NOTE: PLEASE USE THIS DOMAIN NAME IN ALL LOG MESSAGES FROM THIS FILE.
DOMAIN_NAME = "proxy_axis_vio"


class AxisVIOService(services.Service):
    def getName(self) -> str:
        return AXIS_VIO_SERVICE_NAME


class AxisVIOProxy(CorePropertyProxy, AxisVIOService):
    def __init__(self, channel):
        super(AxisVIOProxy, self).__init__(channel)
        self.listeners = {}

    def initialize(self, node_id: str, done: services.DoneHWCommand) -> None:
        log[DOMAIN_NAME].debug("Sending initializeCmd")
        return self.send_xicom_command("initialize", (node_id,), done)

    # ===========================================
    # Core control
    # ===========================================

    def reset_core(
        self, options: typing.Dict[str, typing.Any], done: services.DoneHWCommand
    ) -> None:
        log[DOMAIN_NAME].debug("Sending resetCoreCmd")
        return self.send_xicom_command("resetCore", (options["node_id"],), done)

    # ===========================================
    # Port related
    # ===========================================
    def commit_port_out_data(
        self, node_id: str, options: typing.Dict[str, typing.Any], done: services.DoneHWCommand
    ) -> None:
        log[DOMAIN_NAME].debug(f"Sending commitPortOutDataCmd with data {options['port_data']}")
        return self.send_xicom_command("commitPortOutData", (node_id, options), done)

    def refresh_port_out_data(
        self, options: typing.Dict[str, typing.Any], done: services.DoneHWCommand
    ) -> None:
        log[DOMAIN_NAME].debug(
            f"Sending refreshPortOutDataCmd with data {options['port_out_numbers']}"
        )
        return self.send_xicom_command("refreshPortOutData", (options["node_id"], options), done)

    def refresh_port_in_data(
        self, options: typing.Dict[str, typing.Any], done: services.DoneHWCommand
    ) -> None:
        log[DOMAIN_NAME].debug(f"Sending refreshPortInDataCmd")
        return self.send_xicom_command("refreshPortInData", (options["node_id"],), done)

    def get_port_data(
        self, options: typing.Dict[str, typing.Any], done: services.DoneHWCommand
    ) -> None:
        log[DOMAIN_NAME].debug(f"Sending getPortDataCmd")
        return self.send_xicom_command("getPortData", (options["node_id"],), done)

    # ===========================================
    # Probe related
    # ===========================================

    def define_probe(self, options: typing.Dict[str, typing.Any], done: services.DoneHWCommand):
        log[DOMAIN_NAME].debug(f"Sending defineProbe with data {options['probe_options']}")
        return self.send_xicom_command("defineProbe", (options["node_id"], options), done)

    def report_probe(self, options: typing.Dict[str, typing.Any], done: services.DoneHWCommand):
        log[DOMAIN_NAME].debug(f"Sending reportProbeCmd with data {options['probe_name']}")
        return self.send_xicom_command("reportProbe", (options["node_id"], options), done)

    def undefine_probe(self, options: typing.Dict[str, typing.Any], done: services.DoneHWCommand):
        log[DOMAIN_NAME].debug(f"Sending undefineProbeCmd with data {options['probe_name']}")
        return self.send_xicom_command("undefineProbe", (options["node_id"], options), done)

    def refresh_probe(self, options: typing.Dict[str, typing.Any], done: services.DoneHWCommand):
        log[DOMAIN_NAME].debug(f"Sending refreshProbeCmd with data {options['probe_name']}")
        return self.send_xicom_command("refreshProbe", (options["node_id"], options), done)

    def commit_probe(self, options: typing.Dict[str, typing.Any], done: services.DoneHWCommand):
        log[DOMAIN_NAME].debug(f"Sending commitProbe with data {options['probe_data']}")
        return self.send_xicom_command("commitProbe", (options["node_id"], options), done)



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.10.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# <link rel="preconnect" href="https://fonts.gstatic.com">
# <link href="https://fonts.googleapis.com/css2?family=Fira+Code&display=swap" rel="stylesheet">
#
# ### License
#
# <p style="font-family: 'Fira Code', monospace; font-size: 1.2rem">
# Copyright 2021 Xilinx, Inc.<br><br>
# Licensed under the Apache License, Version 2.0 (the "License");<br>
# you may not use this file except in compliance with the License.<br><br>
# You may obtain a copy of the License at <a href="http://www.apache.org/licenses/LICENSE-2.0"?>http://www.apache.org/licenses/LICENSE-2.0</a><br><br>
# Unless required by applicable law or agreed to in writing, software<br>
# distributed under the License is distributed on an "AS IS" BASIS,<br>
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.<br>
# See the License for the specific language governing permissions and<br>
# limitations under the License.<br>
# </p>
#

# %% [markdown]
# # ChipScoPy DDR 2D Eye Margin Scan Example
#
#
# <img src="../img/api_overview.png" width="500" align="left">

# %% [markdown]
# ## Description
# This demo shows how to exercise and run Versal DDRMC 2D Margin Scan features
#
#
# ## Requirements
# - Local or remote Xilinx Versal board, such as a VCK190
# - Xilinx hw_server 2021.2 installed and running
# - Xilinx cs_server 2021.2 installed and running
# - Python 3.8 or greater installed
# - ChipScoPy 2021.2 installed
# - Jupyter notebook support installed - Please do so, using the command `pip install chipscopy[jupyter]`

# %% [markdown]
# ## 1 - Initialization: Imports and File Paths
#
# After this step,
# - Required functions and classes are imported
# - URL paths are set correctly
# - File paths to example files are set correctly

# %%
import sys
import os
import pprint
from chipscopy import create_session, delete_session, report_versions
from chipscopy import get_design_files
from ddr_scan_util import convert_vref_pct_to_code

# %% pycharm={"is_executing": true}
# Make sure to start the hw_server and cs_server prior to running.
# Specify locations of the running hw_server and cs_server below.
# The default is localhost - but can be other locations on the network.
CS_URL = os.getenv("CS_SERVER_URL", "TCP:localhost:3042")
HW_URL = os.getenv("HW_SERVER_URL", "TCP:localhost:3121")

# The get_design_files() function tries to find the PDI and LTX files. In non-standard
# configurations, you can put the path for PROGRAMMING_FILE and PROBES_FILE below.
design_files = get_design_files("vck190/production/chipscopy_ced")

PROGRAMMING_FILE = design_files.programming_file
PROBES_FILE = design_files.probes_file

print(f"HW_URL: {HW_URL}")
print(f"CS_URL: {CS_URL}")
print(f"PROGRAMMING_FILE: {PROGRAMMING_FILE}")
print(f"PROBES_FILE:{PROBES_FILE}")

# %% pycharm={"is_executing": true}
# Which DDRMC target (0..3) for given ACAP
DDR_INDEX = 0
# Which Rank of the memory interface
RANK = 0
# Read or Write Margin : "READ" "WRITE"
MARGIN_MODE = "READ"
# Data pattern used for margin check : "SIMPLE" "COMPLEX"
DATA_PATTERN = "COMPLEX"
# VREF Percentage Minimum (reccommended: Read :DDR4 25, LP4 5 , Write : DDR4 60  , LP4 10)
VREF_PCT_MIN = 25
# VREF Percentage Maximum (reccommended: Read:DDR4 50 , LP4 35 , Write : DDR4 90  , LP4 30)
VREF_PCT_MAX = 50
# Steps to show in the 2D eye scan  ( 1 step takes ~1 second to capture)
STEPS = 15
# Which nibble (read mode) or byte lane (write) to display
DISPLAY_INDEX = 1

# %% [markdown]
# ## 2 - Create a session and connect to the hw_server and cs_server
#
# The session is a container that keeps track of devices and debug cores.
# After this step,
# - Session is initialized and connected to server(s)
# - Versions are detected and reported to stdout

# %%
session = create_session(cs_server_url=CS_URL, hw_server_url=HW_URL)
report_versions(session)

# %% [markdown]
# ## 3 - Program the device with the example design
#
# After this step,
# - Device is programmed with the example programming file

# %%
# Typical case - one device on the board - get it.
device = session.devices.filter_by(family="versal").get()
device.program(PROGRAMMING_FILE)

# %% [markdown]
# ## 4 - Discover Debug Cores
#
# Debug core discovery initializes the chipscope server debug cores. This brings debug cores in the chipscope server online.
#
# After this step,
#
# - The cs_server is initialized and ready for use

# %%
device.discover_and_setup_cores(ltx_file=PROBES_FILE)
print(f"Debug cores setup and ready for use.")

# %% [markdown]
# ## 5 - Get a list of the integrated DDR Memory Controllers

# %%
ddr_list = device.ddrs
print(f"{len(ddr_list)} integrated DDRMC cores exist on this device.")

ddr_index = 0
for ddr in ddr_list:
    if ddr.is_enabled:
        print(f" DDRMC instance {ddr_index} is enabled")
    else:
        print(f" DDRMC instance {ddr_index} is disabled")
    ddr_index += 1

# %% [markdown]
# ## 6- Select a target DDR by index and display calibration status

# %%
try:
    ddr = ddr_list[DDR_INDEX]
    props = ddr.ddr_node.get_property(["cal_status"])
    print(f"Calibration status of DDRMC instance {DDR_INDEX} is {props['cal_status']}")
except:
    print(f"The DDR controller at index {DDR_INDEX} is not in use")

# %%
## Initialize and activate the Margin Check feature in the DDRMC
ddr.ddr_node.set_property({"mgchk_enable": 1})
ddr.ddr_node.commit_property_group([])
ddr.ddr_node.set_property({"mgchk_enable": 0})
ddr.ddr_node.commit_property_group([])
print("Initialization complete.")

# %% [markdown]
# ## 7 - Setting the 2D eye scan read or write mode

# %%
if MARGIN_MODE == "READ":
    print("Setting 2D eye for READ margin")
    ddr.set_eye_scan_read_mode()
elif MARGIN_MODE == "WRITE":
    print("Setting 2D eye for WRITE margin")
    ddr.set_eye_scan_write_mode()
else:
    print(
        f" ERROR: MARGIN_MODE is set to {MARGIN_MODE} which is an illegal value, only READ or WRITE is allowed"
    )


# %% [markdown]
# ## 8 - Setting the 2D eye scan data pattern mode

# %%
if DATA_PATTERN == "SIMPLE":
    print("Setting 2D eye for SIMPLE data pattern")
    ddr.set_eye_scan_simple_pattern()
elif DATA_PATTERN == "COMPLEX":
    print("Setting 2D eye for COMPLEX data pattern")
    ddr.set_eye_scan_complex_pattern()
else:
    print(
        f" ERROR: DATA_PATTERN is set to {DATA_PATTERN} which is an illegal value, only SIMPLE or COMPLEX is allowed"
    )


# %% [markdown]
# ## 9 - Setting the Vref sample min/max range

# %%
print("Vref Min setting...")
vref_min_code = convert_vref_pct_to_code(ddr, MARGIN_MODE, VREF_PCT_MIN)
print("Vref Max setting...")
vref_max_code = convert_vref_pct_to_code(ddr, MARGIN_MODE, VREF_PCT_MAX)

ddr.set_eye_scan_vref_min(vref_min_code)
ddr.set_eye_scan_vref_max(vref_max_code)
ddr.set_eye_scan_vref_steps(STEPS)
print(f"Dividing the Vref range into {STEPS} steps")


# %% [markdown]
# ## 10 - Run 2D Margin Scan after settings

# %%
ddr.run_eye_scan()

# %% [markdown]
# ## 11 - Display Scan Plots by a given Unit(nibble/byte) index
#
# You can display static or dynamic plots. The display_type controls the display output.
# - "static" is a simple image that can be saved.
# - "dynamic" is an interactive javascript plot.
# - The default is "dynamic".

# %%
ddr.display_eye_scan(DISPLAY_INDEX, display_type="static")

# %% [markdown]
# Optionally you can return figures as a list for later operations.

# %%
figs = ddr.display_eye_scan(DISPLAY_INDEX + 1, return_as_list=True)

# %% [markdown]
# The following loop demonstrates how you can display the graphs from a list created previously.
# It is easy to display interactive or static images.
#
# Here we get the list of figures and output them to png format.

# %%
from IPython.display import Image, display

for fig in figs:
    # To display interactive images, uncomment the following line:
    # fig.show()

    # To display a static png image:
    image_bytes = fig.to_image(format="png")
    ipython_image = Image(image_bytes)
    display(ipython_image)


# %% [markdown]
# ## 12 - Save the Eye Scan data from latest run

# %%
ddr.save_eye_scan_data("myoutput.csv")

# %% [markdown]
# ## 13 - Load Eye Scan data from a given data file

# %%
ddr.load_eye_scan_data("myoutput.csv")

# %% [markdown]
# ## 14 - Review overall Scan status and Control group detail from latest run

# %%
props = ddr.ddr_node.get_property_group(["eye_scan_stat", "eye_scan_ctrl"])
print(pprint.pformat(props, indent=2))

# %% [markdown]
# ## 15 - (Optional) Report Full DDR config and calibration/margin Info

# %%
# (uncomment to see report)
# ddr.report()

# %%
## When done with testing, close the connection
delete_session(session)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
ChipScope Client Overview

Server Connection
^^^^^^^^^^^^^^^^^
.. automethod:: chipscope.client.connect

"""

import threading
from chipscopy.dm import request
from chipscopy.tcf import protocol, channel, peer
from chipscopy.tcf.channel.AbstractChannel import AbstractChannel
from chipscopy.utils.logger import log
from .server_info import ServerInfo
from .util import process_param_str, sync_call, parse_params
from chipscopy import proxies


proxies.init()

_connections = {}
_lock = threading.Lock()


def _setup_channel_listener(params, chan: channel):
    with _lock:
        _connections[params] = chan

    class DisconnectListener(channel.ChannelListener):
        def onChannelClosed(self, error):
            with _lock:
                for peer, c in _connections.items():
                    if c == chan and c.getState() == channel.STATE_CLOSED:
                        log.client.info(f"Channel {peer} closed")
                        _connections.pop(peer)
                        break

    chan.addChannelListener(DisconnectListener())


def connect_channel(params, done: request.DoneCallback = None):
    if not protocol.getEventQueue():
        protocol.startEventQueue()
    wait = ~protocol.isDispatchThread() and not done
    p = peer.TransientPeer(parse_params(process_param_str(params)))
    if wait:
        c = sync_call(_openChannel, p)
    else:
        c = protocol.invokeAndWait(_openChannel, p, done)
    return c


def connect(url: str, done: request.DoneCallback = None) -> ServerInfo:
    """
    Connects to a ChipScope or Hardware server

    :param url: A url in the form of "<hostname or IP>:<port>"
    :param done: Done callback if called asynchronously
    :return: Server Info object
    """

    def done_connect(token, error, result):
        done(token, error, ServerInfo(result))

    if done:
        return connect_channel(url, done_connect)
    return ServerInfo(connect_channel(url))


# @DEPRECATED
def connect_xicom(url):
    url = process_param_str(url)
    c = connect_channel(url)
    return ServerInfo(c)


def disconnect(server):
    chan = None
    if isinstance(server, ServerInfo):
        chan = server.channel
    elif isinstance(server, AbstractChannel):
        chan = server
    if chan:
        protocol.invokeAndWait(chan.close)


def connections():
    with _lock:
        return list(_connections.keys())


def get_channel(peer):
    with _lock:
        return _connections.get(peer)


def _openChannel(p, done=None):
    assert protocol.isDispatchThread()
    log.client.debug(f"Connecting to {p.getID()}")
    c = p.openChannel()
    if done is None:
        return c

    class ChannelListener(channel.ChannelListener):
        def onChannelOpened(self):
            log.client.info(f"Connected to {p.getID()}")
            c.removeChannelListener(self)
            _setup_channel_listener(c.remote_peer.getID(), c)
            done(c, None, c)

        def onChannelClosed(self, error):
            log.client.error(f"Failed to connect to {p.getID()}")
            done(c, error, None)

    c.addChannelListener(ChannelListener())



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

import threading
from chipscopy.utils.logger import log


class EventQueue(object):

    def __init__(self, on_shutdown=None):
        self.__thread = threading.Thread(target=self,
                                         name="TCF Event Dispatcher")
        self.__thread.daemon = True
        self.__is_waiting = False
        self.__is_shutdown = False
        self.__on_shutdown = on_shutdown
        self.__lock = threading.Condition()
        self.__queue = []

    def start(self):
        self.__thread.start()

    def shutdown(self):
        try:
            if self.__on_shutdown:
                self.__on_shutdown()
            with self.__lock:
                self.__is_shutdown = True
                if self.__is_waiting:
                    self.__is_waiting = False
                    self.__lock.notifyAll()
            self.__thread.join()
        except Exception as e:
            from . import protocol  # as protocol import this module too
            protocol.log("Failed to shutdown TCF event dispatch thread", e)

    def isShutdown(self):
        with self.__lock:
            return self.__is_shutdown

    def __error(self, x):
        if not self.isShutdown():
            from . import protocol  # as protocol import this module too
            protocol.log("Unhandled exception in TCF event dispatch", x)

    def __call__(self):
        while True:
            try:
                with self.__lock:
                    while not self.__queue:
                        if self.__is_shutdown:
                            return
                        self.__is_waiting = True
                        self.__lock.wait()
                    r, args, kwargs = self.__queue.pop(0)
                if log.is_domain_enabled("events", "DEBUG"):
                    str_args = ",".join(str(arg) for arg in args)
                    log.events.debug(f"{str(r)}({str_args},{kwargs})")
                r(*args, **kwargs)
            except Exception as x:
                self.__error(x)

    def invokeLater(self, r, *args, **kwargs):
        assert r
        with self.__lock:
            if self.__is_shutdown:
                raise RuntimeError("TCF event dispatcher has shut down")
            self.__queue.append((r, args, kwargs))
            if self.__is_waiting:
                self.__is_waiting = False
                self.__lock.notifyAll()

    def isDispatchThread(self):
        return threading.currentThread() is self.__thread

    def getCongestion(self):
        with self.__lock:
            job_cnt = 0
            l0 = int(job_cnt / 10) - 100
            l1 = int(len(self.__queue) / 10) - 100
            if l1 > l0:
                l0 = l1
            if l0 > 100:
                l0 = 100
            return l0



# *****************************************************************************
# Copyright (c) 2011, 2013, 2016 Wind River Systems, Inc. and others.
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License 2.0
# which accompanies this distribution, and is available at
# https://www.eclipse.org/legal/epl-2.0/
#
# Contributors:
#     Wind River Systems - initial API and implementation
# *****************************************************************************

import threading
from .. import protocol, channel


class Task(object):
    """A <tt>Task</tt> is an utility class that represents the result of an
    asynchronous communication over TCF framework.

    Methods are provided to check if the communication is complete, to wait for
    its completion, and to retrieve the result of the communication.

    Task is useful when communication is requested by a thread other then TCF
    dispatch thread.
    If client has a global state, for example, cached remote data,
    multithreading should be avoided,
    because it is extremely difficult to ensure absence of racing conditions or
    deadlocks in such environment.
    Such clients should consider message driven design, see DataCache and its
    usage as an example.

    If a client is extending Task it should implement run() method to perform
    actual communications.
    The run() method will be execute by TCF dispatch thread, and client code
    should then call either done() or error() to indicate that task
    computations are complete.
    """
    __result = None
    __is_done = False
    __error = None
    __canceled = False
    __channel_listener = None

    def __init__(self, target=None, *args, **kwargs):
        """
        Construct a TCF task object and schedule it for execution.
        """
        if target:
            kwargs["done"] = self.__done
        else:
            target = self.run

        self._target = target
        self._args = args
        self._kwargs = kwargs
        self._lock = threading.Condition()
        self.__channel = kwargs.pop("channel", None)
        protocol.invokeLater(self.__doRun)
        timeout = kwargs.pop("timeout", None)
        if timeout:
            protocol.invokeLaterWithDelay(timeout, self.cancel)

    def __doRun(self):
        try:
            if self.__channel:
                if self.__channel.getState() != channel.STATE_OPEN:
                    raise Exception("Channel is closed")
                task = self

                class CancelOnClose(channel.ChannelListener):
                    def onChannelClosed(self, error):
                        task.cancel(True)
                self.__channel_listener = CancelOnClose()
                self.__channel.addChannelListener(self.__channel_listener)
            self._target(*self._args, **self._kwargs)
        except Exception as x:
            if not self.__is_done and self.__error is None:
                self.error(x)

    def __done(self, error, result):
        if error:
            self.error(error)
        else:
            self.done(result)

    def run(self, *args, **kwargs):
        raise NotImplementedError("Abstract method")

    def done(self, result):
        with self._lock:
            assert protocol.isDispatchThread()
            if self.__canceled:
                return
            assert not self.__is_done
            assert not self.__error
            assert self.__result is None
            self.__result = result
            self.__is_done = True
            if self.__channel:
                self.__channel.removeChannelListener(self.__channel_listener)
            self._lock.notifyAll()

    def error(self, error):
        """
        Set a error and notify all threads waiting for the task to complete.
        The method is supposed to be called in response to executing of run()
        method of this task.

        @param error - computation error.
        """
        assert protocol.isDispatchThread()
        assert error
        with self._lock:
            if self.__canceled:
                return
            assert self.__error is None
            assert self.__result is None
            assert not self.__is_done
            self.__error = error
            if self.__channel:
                self.__channel.removeChannelListener(self.__channel_listener)
            self._lock.notifyAll()

    def cancel(self):
        assert protocol.isDispatchThread()
        with self._lock:
            if self.isDone():
                return False
            self.__canceled = True
            self.__error = Exception("Canceled")
            if self.__channel:
                self.__channel.removeChannelListener(self.__channel_listener)
            self._lock.notifyAll()
        return True

    def get(self, timeout=None):
        """
        Waits if necessary for the computation to complete, and then
        retrieves its result.

        @return the computed result
        @throws CancellationException if the computation was canceled
        @throws ExecutionException if the computation threw an
        exception
        @throws InterruptedException if the current thread was interrupted
        while waiting
        """
        assert not protocol.isDispatchThread()
        with self._lock:
            while not self.isDone():
                self._lock.wait(timeout)
                if timeout and not self.isDone():
                    raise TimeoutException("Timed out")
            if self.__error:
                raise Exception("TCF task aborted", self.__error)
            return self.__result

    def isCancelled(self):
        """
        Returns <tt>true</tt> if this task was canceled before it completed
        normally.

        @return <tt>true</tt> if task was canceled before it completed
        """
        with self._lock:
            return self.__canceled

    def isDone(self):
        """
        Returns <tt>true</tt> if this task completed.

        Completion may be due to normal termination, an exception, or
        cancellation -- in all of these cases, this method will return
        <tt>true</tt>.

        @return <tt>true</tt> if this task completed.
        """
        with self._lock:
            return self.__error or self.__is_done

    def getError(self):
        """
        Return task execution error if any.
        @return Exception object or None
        """
        return self.__error

    def getResult(self):
        return self.__result


class TimeoutException(Exception):
    pass



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

from typing import NewType
import threading
import inspect
import importlib
from .. import protocol
from ..channel import *
from ..channel.Command import Command
from .arguments import *

_providers = []
_lock = threading.RLock()

TokenType = NewType("TokenType", Token)


class TCFError(Exception):
    pass


class DoneHWCommand(object):
    """Client call back interface for generic commands."""
    def doneHW(self, token, error, args):
        """Called when memory operation command command is done.
        :param token: Command handle.
        :param error: Error object or **None**.
        """
        pass

    def __call__(self, token, error, args):
        return self.doneHW(token, error, args)


class ServiceProvider(object):
    """Clients can implement this abstract class if they want to provide
    implementation of a local service or remote service proxy.
    """

    def get_local_service(self, channel):
        pass

    def get_service_proxy(self, channel, service_name):
        pass


def add_service_provider(provider):
    with _lock:
        _providers.append(provider)


def removeServiceProvider(provider):
    with _lock:
        _providers.remove(provider)


def onChannelCreated(channel, services_by_name):
    with _lock:
        # TODO ZeroCopy support is incomplete
        # zero_copy = ZeroCopy()
        # services_by_name[zero_copy.getName()] = zero_copy
        for provider in _providers:
            try:
                arr = provider.get_local_service(channel)
                if not arr:
                    continue
                for service in arr:
                    if service.getName() in services_by_name:
                        continue
                    services_by_name[service.getName()] = service
            except Exception as x:
                protocol.log("Error calling TCF service provider", x)


def onChannelOpened(channel, service_names, services_by_name):
    with _lock:
        for name in service_names:
            for provider in _providers:
                try:
                    service = provider.get_service_proxy(channel, name)
                    if not service:
                        continue
                    services_by_name[name] = service
                    break
                except Exception as x:
                    protocol.log("Error calling TCF service provider", x)
            if name in services_by_name:
                continue
            services_by_name[name] = GenericProxy(channel, name)


def getServiceManagerID():
    # In current implementation ServiceManager is a singleton,
    # so its ID is same as agent ID.
    return protocol.getAgentID()


class GenericCallback(object):

    def __init__(self, callback):
        self.callback = callback

    def __getattr__(self, attr):
        if attr.startswith("done"):
            return self.callback


class Service(object):
    """TCF service base class."""
    def getName(self):
        """Abstract method to get the service name.

        :returns: This service name
        """
        raise NotImplementedError("Abstract method")

    def __init__(self, channel):
        self.channel = channel

    def __str__(self):
        """TCF service string representation.

        :returns: The name of the service.
        """
        return self.getName()

    def _makeCallback(self, done):
        """Turn *done* into a callable.

        If *done* is already a :class:`collections.Callable`, it is returned
        as is, else, it is made callable, and returned.

        :param done: The item to make callable.

        :returns: The callable value of *done*
        """
        if inspect.isfunction(done) or inspect.ismethod(done):
            return GenericCallback(done)
        return done

    def send_command(self, name, args, done):
        done = self._makeCallback(done)
        service = self

        class HWCommand(Command):
            def __init__(self):
                super(HWCommand, self).__init__(service.channel,
                                                service.getName(),
                                                name,
                                                args)

            def done(self, error, results):
                if not error:
                    assert len(results) > 0
                    #error = self.toError(results[0])
                    error = results[0]
                    results = results[1:]
                done.doneHW(self.token, error, results)
        return HWCommand().token

    def send_xicom_command(self, name, args, done, progress=None):
        done = self._makeCallback(done)
        service = self

        class XicomCommand(Command):
            def __init__(self):
                super(XicomCommand, self).__init__(service.channel,
                                                service.getName(),
                                                name,
                                                to_xargs(args))

            def done(self, error, results):
                if not error:
                    assert len(results) >= 1
                    error = results[0]
                    if isinstance(error, dict) and 'Format' in error:
                        e = Exception(error['Format'])
                        try:
                            mod = importlib.import_module(error['Module'])
                            cls = getattr(mod, error["Class"])
                            e = cls(f"TCF Error: {service.getName()} {name}: {error['Format']}")
                        except Exception:
                            pass
                        error = e
                    results = from_xargs(results[1:]) if len(results) > 1 else None
                    if results and len(results) == 1:
                        results = results[0]
                if done:
                    done.doneHW(self.token, error, results)

            def progress_update(self, error, results):
                if not error:
                    assert len(results) >= 1
                    error = results[0]
                    if isinstance(error, dict) and 'Format' in error:
                        e = Exception(error['Format'])
                        try:
                            mod = importlib.import_module(error['Module'])
                            cls = getattr(mod, error["Class"])
                            e = cls(f"TCF Error: {service.getName()} {name}: {error['Format']}")
                        except Exception:
                            pass
                        error = e
                    results = from_xargs(results[1:]) if len(results) > 1 else None
                    if results and len(results) == 1:
                        results = results[0]
                if progress:
                    progress(self.token, error, results)
        return XicomCommand().token

    def send_old_xicom_command(self, name, args, done):
        done = self._makeCallback(done)
        service = self

        class OldXicomCommand(Command):
            def __init__(self):
                super().__init__(service.channel,
                                 service.getName(),
                                 name,
                                 to_xargs(args))

            @staticmethod
            def preprocess_data(data):
                def next_bytes(data):
                    start = data.find(b'(')
                    if start > 0 and data[start-1] == b'"':
                        start = -1
                    return start
                data_bytes = bytearray()
                processed_data = bytearray()
                while len(data):
                    data_tail = b''
                    start = next_bytes(data)
                    while start >= 0:
                        count_stop = data.find(b')',start)
                        count = int(data[start+1:count_stop])
                        stop = count_stop + 1 + count
                        data_bytes += data[count_stop + 1:stop]
                        processed_data += data[:start]
                        processed_data += f"\"({count})\"".encode('utf-8')
                        data = data[stop:]
                        start = next_bytes(data)
                        arg_end = data.find(b'\x00')
                        if 0 <= arg_end < start:
                            data_tail = data[:arg_end]
                            data = data[arg_end+1:]
                            break
                        data_tail = data
                    if data_tail:
                        processed_data += data_tail
                    else:
                        processed_data += data
                        break
                    if data_bytes:
                        processed_data += b'\x00'
                        processed_data += f"({len(data_bytes)})".encode('utf-8') + data_bytes
                    if start < 0:
                        break
                    processed_data += b'\x00'
                    data_bytes.clear()
                return processed_data

            def result(self, token, data):
                data = OldXicomCommand.preprocess_data(data)
                super().result(token, data)

            def done(self, error, results):
                if not error:
                    assert len(results) >= 1
                    error = results[0]
                    if isinstance(error, dict) and 'Format' in error:
                        e = Exception(error['Format'])
                        try:
                            mod = importlib.import_module(error['Module'])
                            cls = getattr(mod, error["Class"])
                            e = cls(f"TCF Error: {service.getName()} {name}: {error['Format']}")
                        except Exception:
                            pass
                        error = e
                    results = from_xargs(results[1:]) if len(results) > 1 else None
                    if results and len(results) == 1:
                        results = results[0]
                if done:
                    done.doneHW(self.token, error, results)
        return OldXicomCommand().token


class ServiceSync(object):
    def __init__(self, channel, service_name, raise_on_error=False):
        self.service_name = service_name
        self.chan = channel
        self.result = None
        self.raise_on_error = raise_on_error

        #services = protocol.invokeAndWait(self.chan.getRemoteServices)
        # if service_name not in services:
        #     raise Exception('No service {} found'.format(service_name))

    def __getattr__(self, attr):
        return CommandWrapper(self, attr)

    def _doneCommand(self, token, error, args):
        self.result = CommandResult(token, error, args)

    def invoke(self, command, *args, **kwargs):
        cond = threading.Condition()
        self.result = None

        def callCommand(command, *args, **kwargs):
            serv = self.chan.getRemoteService(self.service_name)

            class DoneCommand(DoneHWCommand):
                def __init__(self, service):
                    self.service = service

                def doneHW(self, token, error, args):
                    self.service._doneCommand(token, error, args)
                    with cond:
                        cond.notify()

            try:
                kwargs['done'] = DoneCommand(self)
                f = getattr(serv, command, None)
                f(*args, **kwargs)
            except TypeError as e:
                error = '{} service {} error: {}'.format(serv,command,str(e))
                self._doneCommand(Token(),error,[])
                with cond:
                    cond.notify()

        if protocol.isDispatchThread():
            callCommand(command, *args, **kwargs)
        else:
            with cond:
                protocol.invokeLater(callCommand, command, *args, **kwargs)
                # When inside the debugger wait longer, don't time out.
                # todo: move this feature to the command line???
                cond.wait(None if __debug__ else 60)

        if self.raise_on_error and self.result and self.result.error:
            if isinstance(self.result.error, Exception):
                raise self.result.error
            else:
                raise Exception(self.service_name + ':' + command + ': ' + str(self.result.error))
        return self.result


class CommandResult(object):
    def __init__(self, token, error, args):
        self.token = token
        self.error = error
        # unwrap result if only one element
        # if args and len(args) == 1 and \
        #         not isinstance(args, dict):
        #     args = args[0]
        self.args = args

    def __str__(self):
        if self.error:
            return "[%s] error: %s" % (self.token.id, self.error)
        return "[%s] result: %s" % (self.token.id, self.args)
    __repr__ = __str__

    def __iter__(self):
        yield self.error
        yield self.args

    def get(self, key=None):
        if key:
            return self.args.get(key)
        return self.args


class CommandWrapper(object):
    def __init__(self, control, command):
        self._control = control
        self._command = command

    def __call__(self, *args, **kwargs):
        return self._control.invoke(self._command, *args, **kwargs)


class ZeroCopy(Service):
    def getName(self):
        return "ZeroCopy"


class GenericProxy(Service):
    """Objects of GenericProxy class represent remote services, which don't
    have a proxy class defined for them.

    Clients still can use such services, but framework will not provide
    service specific utility methods for message formatting and parsing.
    """

    def __init__(self, channel, name):
        self.__channel = channel
        self.name = name

    def getName(self):
        return self.name

    def getChannel(self):
        return self.__channel


class DefaultServiceProvider(ServiceProvider):
    package_base = str(__package__) + ".remote"

    def get_local_service(self, channel):
        # TODO DiagnosticsService
        # return [DiagnosticsService(channel)]
        return []

    def get_service_proxy(self, channel, service_name):
        package_base = str(__package__) + ".remote"
        service = None
        try:
            # Xicom_v1.XX special case
            if 'Xicom_v1' in service_name:
                service_name = "Xicom"
            clsName = service_name + "Proxy"
            clsName = clsName.replace('.', '_')
            package = package_base + "." + clsName
            clsModule = __import__(package, fromlist=[clsName],
                                   globals=globals())
            cls = clsModule.__dict__.get(clsName)
            service = cls(channel)
            # assert service_name == service.getName()
        except ImportError:
            pass
        except Exception as x:
            protocol.log("Cannot instantiate service proxy for " +
                         service_name, x)
        return service


add_service_provider(DefaultServiceProvider())



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
import json
import sys
from io import TextIOBase
from dataclasses import dataclass, asdict
from pprint import pformat
from typing import Dict, List, Union, Optional
from chipscopy.dm import request
from chipscopy.shared.ila_util import to_bin_str
from chipscopy.api._detail.ltx import Ltx, LtxStreamRef
from chipscopy.client.axis_ila_core_client import AxisIlaCoreClient as TCF_AxisIlaCoreClient
from chipscopy.api.ila import (
    ILACaptureCondition,
    ILATriggerCondition,
    ILATrigInMode,
    ILATrigOutMode,
    ILAControl,
    ILA_TRIGGER_POSITION_HALF,
    ILA_WINDOW_SIZE_MAX,
    ILAStatus,
    ILA_STATUS_MEMBERS,
)
from chipscopy.api import CoreInfo, dataclass_fields, filter_props
from chipscopy.api.ila import ILAPort, ILAProbe, ILAProbeValues, ILABitRange
from chipscopy.api.ila import ILAWaveformProbe, ILAWaveform
from chipscopy.api.ila.ila_capture import (
    tcf_refresh_status,
    post_process_status,
    control_from_tcf,
    control_to_tcf,
    tcf_props_to_status,
)
from chipscopy.api.ila.ila_probe import (
    ports_from_tcf_props,
    verify_ports,
    verify_probe_value,
    create_probes_from_ports_and_ltx,
    ILAProbeRadix,
    verify_probe_enum_def,
)
from chipscopy.api.ila.ila_waveform import (
    export_waveform_to_stream,
    get_waveform_data_values,
    tcf_get_waveform_data,
)
from chipscopy.api._detail.debug_core import DebugCore
from chipscopy import CoreType
from chipscopy.shared.ila_data import RAW_DATA_SAMPLE_ALIGN_BIT_COUNT_32
from chipscopy.utils import Enum2StrEncoder


@dataclass(frozen=True)
class ILAStaticInfo:
    """Feature set and dimensions of the ILA core."""

    data_depth: int
    """Data depth in samples."""
    data_width: int
    """Waveform data sample width, in bits."""
    has_advanced_trigger: bool
    """ILA has advanced trigger mode."""
    has_capture_control: bool
    """ILA has basic capture control."""
    has_trig_in: bool
    """ILA has TRIG-IN port."""
    has_trig_out: bool
    """ILA has TRIG-OUT port."""
    match_unit_count: int
    """Total number of probe compare match units."""
    port_count: int
    """Number of probe ports."""

    def __str__(self) -> str:
        return pformat(self.__dict__, 2)


ILA_STATIC_INFO_MEMBERS = dataclass_fields(ILAStaticInfo)


class ILA(DebugCore["AxisIlaCoreClient"]):
    """
    Data read-only attributes are accessible by dot-notation', e.g.  *my_ila.static_info.status*

    ============= =========================================== ===========================================
    Attribute     Type                                        Description
    ============= =========================================== ===========================================
    name          str                                         Name of ILA core instance.
    core_info     :class:`~chipscopy.api.CoreInfo`            Debug core version information.
    static_info   :class:`.ILAStaticInfo`                     Feature set and dimensions of the ILA core.
    status        :class:`.ILAStatus`                         Dynamic status of ILA capture controller.
    control       :class:`.ILAControl`                        Trigger setup.
    ports         [:class:`.ILAPort`]                         List of ILA core ports.
    probes        {str, :class:`.ILAProbe`}                   Probes by probe name.
    probe_values  {str, :class:`.ILAProbeValues`}             Probe compare values.
    waveform      :class:`.ILAWaveform`                       Last uploaded waveform data, or None.
    ============= =========================================== ===========================================

    """

    __next_ila_number = 1  # simple incrementing number tracking - hw_ila_1, hw_ila_2, etc

    def __init__(self, tcf_ila: TCF_AxisIlaCoreClient, device, ltx: Ltx):

        DebugCore.__init__(self, core_type=CoreType.AXIS_ILA, core_tcf_node=tcf_ila)
        init_vals = ILA._init(tcf_ila, self.core_info, ltx)

        self.static_info: ILAStaticInfo = init_vals["static_info"]
        self.status: ILAStatus = init_vals["status"]
        self.control: ILAControl = init_vals["control"]
        self.name: str = init_vals["name"]
        self.ports: [ILAPort] = init_vals["ports"]
        self.probes: {str, ILAProbe} = init_vals["probes"]
        self.probe_values: {str, ILAProbeValues} = init_vals["probe_values"]
        self.waveform: Optional[ILAWaveform] = None
        self._probe_to_port_seqs: {str, [ILABitRange]} = init_vals["probe_to_port_seqs"]
        self._device = device
        self._downstreams_refs: [LtxStreamRef] = ltx.get_downstream_refs(self.name) if ltx else []
        self.raw_sample_align_bit_count = RAW_DATA_SAMPLE_ALIGN_BIT_COUNT_32

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "uuid": self.core_info.uuid}

    def __str__(self) -> str:
        """Returns instance name"""
        return self.name

    def __repr__(self) -> str:
        ret_dict = {
            "name": self.name,
            "core_info": asdict(self.core_info),
            "static_info": asdict(self.static_info),
            "status": asdict(self.status),
            "control": asdict(self.control),
            "ports": [asdict(port) for port in self.ports],
            "probes": {key: asdict(val) for key, val in self.probes.items()},
            "probe_values": {key: asdict(val) for key, val in self.probe_values.items()},
        }
        json_dict = json.dumps(ret_dict, cls=Enum2StrEncoder, indent=4)
        return json_dict

    def refresh_status(self) -> None:
        """Read dynamic status and store in attribute 'ila.status'."""
        self.status = tcf_refresh_status(self.core_tcf_node)

    def run_trigger_immediately(
        self,
        trigger_position: int = 0,
        window_count: int = 1,
        window_size: int = ILA_WINDOW_SIZE_MAX,
        trig_out: ILATrigOutMode = ILATrigOutMode.DISABLED,
    ) -> None:
        """Trigger ILA immediately.

        Args:
            trigger_position (int): Default : 0
              Range [0..window_size-1].
            window_count (int): Number of windows to capture. Default value: 1.
            window_size (int): Number of samples per window. Must be a power-of-two value.
               Default value: :attr:`.ILA_WINDOW_SIZE_MAX`
            trig_out (ILATrigOutMode): Specify what drives TRIG-OUT port.
               Default value: :attr:`.ILATrigOutMode.DISABLED`
        """
        self.run_basic_trigger(
            trigger_position,
            window_count,
            window_size,
            ILATriggerCondition.IMMEDIATELY,
            ILACaptureCondition.ALWAYS,
            ILATrigInMode.DISABLED,
            trig_out,
        )

    def run_basic_trigger(
        self,
        trigger_position: int = ILA_TRIGGER_POSITION_HALF,
        window_count: int = 1,
        window_size: int = ILA_WINDOW_SIZE_MAX,
        trigger_condition: ILATriggerCondition = ILATriggerCondition.AND,
        capture_condition: ILACaptureCondition = ILACaptureCondition.ALWAYS,
        trig_in: ILATrigInMode = ILATrigInMode.DISABLED,
        trig_out: ILATrigOutMode = ILATrigOutMode.DISABLED,
    ) -> None:
        """Trigger using probe compare values.

        Args:
            trigger_position (int): Default :attr:`~.ILA_TRIGGER_POSITION_HALF`
                 denotes the middle of the window. Range [0..window_size-1].
            window_count (int): Number of windows to capture. Default value: 1.
            window_size (int): Number of samples per window. Must be a power-of-two value.
                 Default value: :attr:`.ILA_WINDOW_SIZE_MAX`

            trigger_condition (ILATriggerCondition): Trigger condition global boolean operator.
            capture_condition (ILACaptureCondition): Capture condition global boolean operator,
                 to filter samples.
            trig_in (ILATrigInMode): Usage of TRIG-IN Port.
                 Default value: :attr:`.ILATrigInMode.DISABLED`
            trig_out (ILATrigOutMode): Specify what drives TRIG-OUT port.
               Default value: :attr:`ILATrigOutMode.DISABLED`

        """

        def to_tcf_trigger_value(values: [], bit_width: int, enum_def: enum.EnumMeta) -> [str]:
            it = iter(values)
            res = [op + to_bin_str(val, bit_width, enum_def) for op, val in zip(it, it)]
            return res

        def probe_values_to_dict(probe_values: {str, ILAProbeValues}) -> {}:
            res = {}
            for name, p_values in probe_values.items():
                vals = {}
                if p_values.trigger_value:
                    vals["trigger_value"] = to_tcf_trigger_value(
                        p_values.trigger_value, p_values.bit_width, p_values.enum_def
                    )
                if self.static_info.has_capture_control and p_values.capture_value:
                    vals["capture_value"] = to_tcf_trigger_value(
                        p_values.capture_value, p_values.bit_width, p_values.enum_def
                    )
                if vals:
                    res[name] = vals
            return res

        # Window size must be an exponent-of-two.
        w_size = (
            self.static_info.data_depth // window_count
            if window_size == ILA_WINDOW_SIZE_MAX
            else window_size
        )
        t_pos = w_size // 2 if trigger_position == ILA_TRIGGER_POSITION_HALF else trigger_position
        control: ILAControl = ILAControl(
            capture_condition, trig_in, trig_out, trigger_condition, t_pos, "", window_count, w_size
        )

        self.control = control
        # Set probe trigger values.
        self.core_tcf_node.reset_probe(reset_trigger_values=True, reset_capture_values=True)
        tcf_probe_values = probe_values_to_dict(self.probe_values)
        if tcf_probe_values:
            self.core_tcf_node.set_probe(tcf_probe_values)

        control_props = control_to_tcf(self.control)
        self.core_tcf_node.set_property(control_props)
        self._arm2()
        self.core_tcf_node.arm()
        self.waveform = None

    def _arm2(self):
        # NYI feature
        pass

    def run_advanced_trigger(
        self,
        trigger_state_machine: str,
        trigger_position: int = ILA_TRIGGER_POSITION_HALF,
        window_count: int = 1,
        window_size: int = ILA_WINDOW_SIZE_MAX,
        capture_condition: ILACaptureCondition = ILACaptureCondition.ALWAYS,
        trig_out: ILATrigOutMode = ILATrigOutMode.DISABLED,
    ) -> None:
        """ NYI"""

    def upload(self) -> bool:
        """
        Upload waveform. If the ILA has not triggered yet, it will return with value *False*.
        The uploaded waveform is at *self.waveform*.

        Returns (bool):
            True, if a waveform was uploaded.
        """

        self.waveform = None
        uploaded = self.core_tcf_node.upload()
        if uploaded:
            wave = tcf_get_waveform_data(self.core_tcf_node)
            wave["probes"] = self._make_waveform_probes()
            self.waveform = ILAWaveform(**wave)
        return uploaded

    def wait_till_done(self, max_wait_minutes: float = None) -> ILAStatus:
        """
        Wait until all data has been captured, or until timeout.
        Call this function after arming the ILA, before uploading the waveform.

        Args:
            max_wait_minutes (float): Max time in minutes for command. If *None*, status monitor never times out.

        Returns (:class:`.ILAStatus`): ILA capture status.
        """
        return self.monitor_status(max_wait_minutes)

    def monitor_status(
        self, max_wait_minutes: float = None, progress=None, done: request.DoneFutureCallback = None
    ) -> ILAStatus or request.CsFutureRequestSync:
        """
        Function monitors ILA capture status and waits until all data has been captured, or until timeout
        or the function is cancelled.
        Call this function after arming the ILA, before uploading the waveform.
        The command operates in synchronous mode if *done* argument has default value *None* .

        **Synchronous Mode**

        - Calling this function in synchronous mode, is the same as calling function ILA.wait_until_done().
        - Function waits until all data has been captured in the ILA core, or timeout.
        - Use default argument value *None* for arguments *progress* and *done* .
        - Returns an ILAStatus object.

        **Asynchronous Mode**

        - This mode is useful for reporting the capture status to stdout or a GUI.
        - Function does not block. The main thread continues with the next statement.
        - Returns a *future* object, which represents the monitor.
        - A blocking function should be called later on the *future* object, in order to wait on the calling
          thread until the status monitor has completed.
        - Asynchronous Mode is selected, by specifying a *done* function, which is called after the function has completed.
        - If no user defined callback is needed, set *done* argument to dummy function *chipscopy.null_callback* ,
          to enable asynchronous mode.

        **Future Object**

        When the *monitor_status* function is called in asynchronous mode, it will return a *future* object.
        The *future* object has blocking attributes and functions, which will block the current thread until
        the status monitor has completed.

        Blocking Attributes:

        - *future.result* (None or ILAStatus) - ILAStatus object if capture completed successfully, without timeout.
        - *future.error* - (None or Exception) - None if no error otherwise an exception object, e.g. timeout exception.

        Non-blocking Attribute:

        - *future.progress* (None or ILAStatus) - Access in *progress function*, to read the ILA capture status.

        Blocking Function:

        - *future.wait(timeout=None)* - *None* means wait until status monitor completed. Argument *timeout* is in seconds.

        Non-blocking Function:

        - *future.cancel()*  - Cancels the status monitor. An exception will be raised, in the thread which
          called the function *ILA.monitor_status* .

        **Progress Function**

        - User-defined function which takes one argument *future* .
        - Useful for reporting ILA capture status.
        - The *progress function* is only called when the ILA capture status changes.
        - The *progress function* is called in the TCF Event Thread. The *progress function* must not call any
          ChipScoPy API function, which interacts with the cs_server or the device.

        Example:
        ::

             def monitor_status_done(future):
                if not future.error:
                    # future.result holds an ILAStatus object.
                    print_status(future.progress)


        Args:
            max_wait_minutes (float): Max time in minutes for status monitor. If *None*, status monitor never times out.
            progress(progress_fn) : See Asynchronous Mode, above. This function runs in the TCF Event Thread.
            done(request.DoneFutureCallback): Done callback. This function runs in the TCF Event Thread.

        Returns (:class:`.ILAStatus` or request.CsFutureRequestSync or None):
        """

        def final(future):  # called on main thread
            if future._result:
                self.status = future._result

        if progress and not done:
            raise ValueError(
                "Function ila.monitor_status(): Cannot specify *progress* argument when argument *done* is None."
            )

        future = self.core_tcf_node.future(done=done, final=final, progress=progress)
        return future.monitor_status(tcf_props_to_status, max_wait_minutes=max_wait_minutes)

    def reset_probes(
        self, reset_trigger_values: bool = True, reset_capture_values: bool = True
    ) -> None:
        """Reset probe compare values, to dont-cares.

        Args:
           reset_trigger_values (bool): Resets trigger compare values, for all core probes.
           reset_capture_values (bool): Resets basic capture control compare value, for all probes.
        """
        if reset_trigger_values:
            for values in self.probe_values.values():
                values.trigger_value = []

        if reset_capture_values:
            for values in self.probe_values.values():
                values.capture_value = []

    def set_probe_trigger_value(self, name: str, trigger_value: []) -> None:
        """For basic trigger mode, set probe compare value(s).

            A compare value consists of a pair: <operator> and <value>
            <value> may be of type int or binary string of the correct bit_width.
            An empty list is a don't-care value.

                Example: Test for LSB is '0' on a 4-bit probe.
                   ['==',   'XXX0']

                Example: Range check, assuming the global Trigger Condition is ILATriggerCondition.AND:
                   ['>=',   '0011',   '<=',   '1010']

                Example: Range check, assuming the global Trigger Condition is ILATriggerCondition.AND:
                   ['>=',   3,   '<=',   10]

                Example: Test equal to any of 3 numbers, assuming the global Trigger Condition is ILATriggerCondition.OR:
                   ['==',   '0011',   '==',   '0111',   '==',   '1111']

                Example: 4-bit dont-care value.
                   ['==',   'XXXX']

                Example: Dont-care value, written as an empty list.
                   []

            Some probes have associated enum values, which can be used when setting values.

                Example: Use an Enum value.
                    ['==', Colors.BLUE]

                Example: Use Enum name string.
                    ['==', "BLUE"]

        Args:
           name (str): Probe name.
           trigger_value (list): [<operator>, <value>, ...]   See :class:`~.ILAProbeValues`
        """
        self._set_probe_value(name, trigger_value, is_trigger=True)

    def set_probe_capture_value(self, name: str, capture_value: []) -> None:
        """For basic capture mode, set probe compare value to filter samples.

        Args:
           name (str): Probe name.
           capture_value: List with two items: [<operator>, <value>].   See :class:`~.ILAProbeValues`
        """
        if not self.static_info.has_capture_control:
            raise ValueError(
                f'Cannot set capture value for Probe "{name}", '
                f'since ILA "{self.name}" does not support basic capture control.'
            )
        self._set_probe_value(name, capture_value, is_trigger=False)

    def get_probe_trigger_value(self, name: str) -> []:
        """Get basic trigger mode compare value(s) for a probe.

        Args:
           name (str): Probe name.

        returns:
           List of <operator>/<value> pairs.  See :class:`~.ila_probe.ILAProbeValues`
        """
        values = self.probe_values.get(name, None)
        if not values:
            raise KeyError(f"Probe {name} is not a defined trigger probe.")
        return values.trigger_value

    def get_probe_capture_value(self, name: str) -> []:
        """Get basic capture mode compare value for a probe.

        Args:
           name (str): Probe name.

        returns:
           List with two items [<operator>, <value>].   See :class:`~.ILAProbeValues`
        """
        values = self.probe_values.get(name, None)
        if not values:
            raise KeyError(f"Probe {name} is not a defined trigger probe.")
        return values.capture_value

    def _set_probe_value(self, name: str, value: [], is_trigger: bool) -> None:
        probe = self.probes.get(name, None)
        values: ILAProbeValues = self.probe_values.get(name, None)
        if not probe or not values:
            raise KeyError(f"Probe {name} is not a defined trigger probe.")
        verify_probe_value(value, is_trigger, probe, values.enum_def)
        if is_trigger:
            values.trigger_value = value
        else:
            values.capture_value = value

    @staticmethod
    def _init(tcf_ila: TCF_AxisIlaCoreClient, core_info: CoreInfo, ltx: Optional[Ltx]):
        ila_name = f"hw_ila_{ILA.__next_ila_number}"
        ILA.__next_ila_number += 1
        props = tcf_ila.get_property_group(["static_info", "status", "control"])
        post_process_status(props)

        static_info = ILAStaticInfo(**filter_props(props, ILA_STATIC_INFO_MEMBERS))
        status = ILAStatus(**filter_props(props, ILA_STATUS_MEMBERS))
        control = control_from_tcf(props)
        ports = ports_from_tcf_props(props)
        if ltx:
            verify_ports(ltx, core_info.uuid, ports)

        probes, probe_to_port_seqs, cell_name, enum_defs = create_probes_from_ports_and_ltx(
            tcf_ila, ports, ltx, core_info.uuid
        )
        probe_values = {
            probe.name: ILAProbeValues(
                probe.bit_width,
                [],
                [],
                ILAProbeRadix.ENUM if probe.name in enum_defs else ILAProbeRadix.HEX,
                enum_defs.get(probe.name, None),
            )
            for probe in probes.values()
            if probe.is_trigger
        }
        if cell_name:
            ila_name = cell_name
        return {
            "core_info": core_info,
            "static_info": static_info,
            "status": status,
            "control": control,
            "name": ila_name,
            "ports": ports,
            "probes": probes,
            "probe_to_port_seqs": probe_to_port_seqs,
            "probe_values": probe_values,
        }

    def get_probe_enum(self, name: str) -> Optional[enum.EnumMeta]:
        """Get enum class, which defines valid enum values for the probe.

        Args:
           name (str): Probe name.

        Returns:
            Enum class defining the enum values or None.
        """
        probe = self.probes.get(name, None)
        values: ILAProbeValues = self.probe_values.get(name, None)
        if not probe or not values:
            raise KeyError(f"Probe {name} is not a defined probe.")
        return values.enum_def

    def set_probe_enum(
        self, name: str, enum_def: enum.EnumMeta, display_as_enum: bool = True
    ) -> None:
        """Set enum values for a probe, by providing a enum class.

        Args:
           name (str): Probe name.
           enum_def (enum.EnumMeta): Enum class defining enum values, e.g. an enum.IntEnum class.
           display_as_enum (bool): If true, sets probe display index to ``ILAProbeRadix.ENUM``
        """

        probe = self.probes.get(name, None)
        values: ILAProbeValues = self.probe_values.get(name, None)
        if not probe or not values:
            raise KeyError(f"Probe {name} is not a defined probe.")
        verify_probe_enum_def(probe, enum_def)
        values.enum_def = enum_def
        if display_as_enum:
            values.display_radix = ILAProbeRadix.ENUM

    def _make_waveform_probes(self) -> {str, ILAWaveformProbe}:
        return {
            p.name: ILAWaveformProbe(
                p.name,
                p.map,
                self._probe_to_port_seqs[p.map],
                p.is_bus,
                p.bus_left_index,
                p.bus_right_index,
                self.probe_values[p.name].display_radix,
                self.probe_values[p.name].enum_def,
            )
            for p in self.probes.values()
            if p.is_data
        }


def export_waveform(
    waveform: ILAWaveform,
    export_format: str = "CSV",
    fh_or_filepath: Union[TextIOBase, str] = sys.stdout,
    probe_names: Optional[List[str]] = None,
    start_window_idx: int = 0,
    window_count: Optional[int] = None,
    start_sample_idx: int = 0,
    sample_count: Optional[int] = None,
) -> None:
    """
    Export a waveform in VCD or CSV format, to a file.
    By default all samples for all probes are exported,
    but it is possible to select which probes and window/sample ranges.

    Args:
        waveform (ILAWaveform): waveform data.
        export_format (str):  Alternatives for output format.

            - 'CSV' - Comma Separated Value Format. Default.
            - 'VCD' - Value Change Dump.

        fh_or_filepath (TextIOBase, str): File object handle or filepath string. Default is sys.stdout.
            If a file object, the file object is opened and closed by the caller, if needed.

            Alternatively, a file path may be specified as a string.
            The file will be opened and closed by the function.
        probe_names (Optional[List[str]]): List of probe names. Default 'None' means export all probes.
        start_window_idx (int): Starting window index. Default is first window.
        window_count (Optional[int]): Number of windows to export. Default is all windows.
        start_sample_idx (int): Starting sample within window. Default is first sample.
        sample_count (Optional[int]): Number of samples per window. Default is all samples.

    """
    if not waveform:
        raise TypeError('Function export_waveform() called with argument "waveform" set to "None".')

    if isinstance(fh_or_filepath, str):
        with open(fh_or_filepath, "w", buffering=16384) as fh:
            export_waveform_to_stream(
                waveform,
                export_format,
                fh,
                probe_names,
                start_window_idx,
                window_count,
                start_sample_idx,
                sample_count,
            )
        return
    else:
        export_waveform_to_stream(
            waveform,
            export_format,
            fh_or_filepath,
            probe_names,
            start_window_idx,
            window_count,
            start_sample_idx,
            sample_count,
        )


def get_waveform_data(
    waveform: ILAWaveform,
    probe_names: Optional[List[str]] = None,
    start_window_idx: int = 0,
    window_count: Optional[int] = None,
    start_sample_idx: int = 0,
    sample_count: Optional[int] = None,
    include_trigger: bool = False,
    include_sample_info: bool = False,
) -> Dict[str, List[int]]:
    """
    Get probe waveform data as a list of int values for each probe.
    By default all samples for all probes are included in return data,
    but it is possible to select which probes and window/sample ranges.

    Args:
        waveform (ILAWaveform): waveform data.
        probe_names (Optional[List[str]]): List of probe names. Default 'None' means export all probes.
        start_window_idx (int): Starting window index. Default is first window.
        window_count (Optional[int]): Number of windows to export. Default is all windows.
        start_sample_idx (int): Starting sample within window. Default is first sample.
        sample_count (Optional[int]): Number of samples per window. Default is all samples.
        include_trigger (bool): Include pseudo probe with name '__TRIGGER' in result. Default is False.
        include_sample_info (bool):  Default is False. Include the following pseudo probes in result:

          - '__SAMPLE_INDEX' - Sample index
          - '__WINDOW_INDEX' - Window index.
          - '__WINDOW_SAMPLE_INDEX' - Sample index within window.

    Returns (Dict[str, List[int]]):
        Ordered dict, in order:
          - '__TRIGGER', if argument **include_trigger** is True
          - '__SAMPLE_INDEX', if argument **include_sample_info** is True
          - '__WINDOW_INDEX', if argument **include_sample_info** is True
          - '__WINDOW_SAMPLE_INDEX', if argument **include_sample_info** is True
          - probe values in order of argument **probe_names**.

        Dict key: probe name. Dict value is list of int values, for a probe.

    """
    if not waveform:
        raise TypeError(
            'Function get_waveform_data() called with argument "waveform" set to "None".'
        )

    return get_waveform_data_values(
        waveform,
        probe_names,
        start_window_idx,
        window_count,
        start_sample_idx,
        sample_count,
        include_trigger,
        include_sample_info,
    )


def get_waveform_probe_data(
    waveform: ILAWaveform,
    probe_name: str,
    start_window_idx: int = 0,
    window_count: Optional[int] = None,
    start_sample_idx: int = 0,
    sample_count: Optional[int] = None,
) -> List[int]:
    """
    Get waveform data as a list of int values for one probe.
    By default all samples for the probe are returned,
    It is possible to select window range and sample range.

    Args:
        waveform (ILAWaveform): waveform data.
        probe_name (str): probe name.
        start_window_idx (int): Starting window index. Default is first window.
        window_count (Optional[int]): Number of windows to export. Default is all windows.
        start_sample_idx (int): Starting sample within window. Default is first sample.
        sample_count (Optional[int]): Number of samples per window. Default is all samples.

    Returns (List[int]):
        List probe values.

    """
    if not waveform:
        raise TypeError(
            'Function get_waveform_probe_data() called with argument "waveform" set to "None".'
        )

    res_dict = get_waveform_data_values(
        waveform,
        [probe_name],
        start_window_idx,
        window_count,
        start_sample_idx,
        sample_count,
        include_trigger=False,
        include_sample_info=False,
    )
    return res_dict[probe_name]



# *****************************************************************************
# * Copyright (c) 2011, 2013, 2015-2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

import sys
import threading
import time

from .. import compat, protocol, transport, services, peer, errors
from ..services import locator
from ..channel import STATE_CLOSED, STATE_OPEN, STATE_OPENING
from ..channel import Token, fromJSONSequence, toJSONSequence, ChannelListener

EOS = -1  # End Of Stream
EOM = -2  # End Of Message


class Message(object):
    def __init__(self, typeCode):
        if isinstance(typeCode, int):
            typeCode = chr(typeCode)
        self.type = typeCode
        self.service = None
        self.name = None
        self.data = None
        self.is_canceled = None
        self.is_sent = None
        self.token = None
        self.trace = ()

    def __str__(self):
        return "%s %s %s %s" % (self.type, str(self.token), self.service, self.name)


class ReaderThread(threading.Thread):
    def __init__(self, channel, handleInput):
        super(ReaderThread, self).__init__(name="TCF Reader Thread")
        self.channel = channel
        self.handleInput = handleInput
        self.buf = bytearray()
        self.eos_err_report = None
        self.daemon = True

    def error(self):
        raise IOError("Protocol syntax error")

    def readBytes(self, end, buf=None):
        if buf is None:
            buf = bytearray()
        while True:
            n = self.channel.read()
            if n <= 0:
                if n == end:
                    break
                if n == EOM:
                    raise IOError("Unexpected end of message")
                if n < 0:
                    raise IOError("Communication channel is closed by " +
                                  "remote peer")
            buf.append(n)
        return buf

    def readString(self):
        del self.buf[:]
        return self.readBytes(0, self.buf).decode("UTF8")

    def run(self):
        try:
            while True:
                n = self.channel.read()
                if n == EOM:
                    continue
                if n == EOS:
                    try:
                        self.eos_err_report = self.readBytes(EOM)
                        reportLen = len(self.eos_err_report)
                        if reportLen == 0 or reportLen == 1 and \
                           self.eos_err_report[0] == 0:
                            self.eos_err_report = None
                    except:
                        pass
                    break
                msg = Message(n)
                if self.channel.read() != 0:
                    self.error()
                typeCode = msg.type
                if typeCode == 'C':
                    msg.token = Token(self.readBytes(0))
                    msg.service = self.readString()
                    msg.name = self.readString()
                    msg.data = self.readBytes(EOM)
                elif typeCode in 'PRN':
                    msg.token = Token(self.readBytes(0))
                    msg.data = self.readBytes(EOM)
                elif typeCode == 'E':
                    msg.service = self.readString()
                    msg.name = self.readString()
                    msg.data = self.readBytes(EOM)
                elif typeCode == 'F':
                    msg.data = self.readBytes(EOM)
                else:
                    self.error()
                protocol.invokeLater(self.handleInput, msg)
                delay = self.channel.local_congestion_level
                if delay > 0:
                    time.sleep(delay / 1000.0)
            protocol.invokeLater(self.handleEOS)
        except Exception as x:
            try:
                x.tb = sys.exc_info()[2]
                protocol.invokeLater(self.channel.terminate, x)
            except:
                # TCF event dispatcher has shut down
                pass

    def handleEOS(self):
        if not self.channel.out_tokens and not self.eos_err_report:
            self.channel.close()
        else:
            x = IOError("Communication channel is closed by remote peer")
            if self.eos_err_report:
                try:
                    args = fromJSONSequence(self.eos_err_report)
                    if len(args) > 0 and args[0] is not None:
                        x.caused_by = Exception(errors.toErrorString(args[0]))
                except IOError:
                    pass
            self.channel.terminate(x)


class AbstractChannel(object):
    """
    AbstractChannel implements communication link connecting two end points
    (peers). The channel asynchronously transmits messages: commands, results
    and events. Clients can subclass AbstractChannel to support particular
    transport (wire) protocol.

    .. see also: see StreamChannel for stream oriented transport protocols.
    """

    def __init__(self, remote_peer, local_peer=None):
        self.remote_peer = remote_peer
        self.local_peer = local_peer  # TODO
        self.inp_thread = ReaderThread(self, self.__handleInput)
        self.out_thread = threading.Thread(target=self.__write_output,
                                           name="TCF Channel Transmitter")
        self.out_thread.daemon = True
        self.out_tokens = {}
        self.out_queue = []
        self.out_lock = threading.Condition()
        self.pending_command_limit = 32
        self.remote_service_by_class = {}
        self.local_service_by_name = {}
        self.remote_service_by_name = {}
        self.channel_listeners = []
        self.event_listeners = {}
        self.command_servers = {}
        self.redirect_queue = []
        self.redirect_command = None
        self.notifying_channel_opened = False
        self.registered_with_trasport = False
        self.state = STATE_OPENING
        self.proxy = None
        self.zero_copy = False

        self.local_congestion_level = -100
        self.remote_congestion_level = -100
        self.local_congestion_cnt = 0
        self.local_congestion_time = 0
        self.local_service_by_class = {}
        self.trace_listeners = []
        self.linked_channels = {}

    def add_linked_channel(self, other_channel):
        channel = self

        class CloseListener(ChannelListener):
            def onChannelClosed(self, error):
                try:
                    del channel.linked_channels[other_channel]
                except KeyError:
                    return
                if not channel.linked_channels:
                    channel.close()

        if other_channel not in self.linked_channels:
            self.linked_channels[other_channel] = CloseListener()
            other_channel.addChannelListener(self.linked_channels[other_channel])

    @property
    def is_closed(self):
        return self.state == STATE_CLOSED

    def __write_output(self):
        try:
            while True:
                msg = None
                last = False
                with self.out_lock:
                    while len(self.out_queue) == 0:
                        self.out_lock.wait()
                    msg = self.out_queue.pop(0)
                    if not msg:
                        break
                    last = len(self.out_queue) == 0
                    if msg.is_canceled:
                        if last:
                            self.flush()
                        continue
                    msg.is_sent = True
                if msg.trace:
                    protocol.invokeLater(self.__traceMessageSent, msg)
                self.write(msg.type)
                self.write(0)
                if msg.token:
                    self.write(msg.token.id)
                    self.write(0)
                if msg.service:
                    self.write(msg.service.encode("UTF8"))
                    self.write(0)
                if msg.name:
                    self.write(msg.name.encode("UTF8"))
                    self.write(0)
                if msg.data:
                    self.write(msg.data)
                self.write(EOM)
                delay = 0
                level = self.remote_congestion_level
                if level > 0:
                    delay = level * 10
                if last or delay > 0:
                    self.flush()
                if delay > 0:
                    time.sleep(delay / 1000.0)
                # else yield()
            self.write(EOS)
            self.write(EOM)
            self.flush()
        except Exception as x:
            try:
                protocol.invokeLater(self.terminate, x)
            except:
                # TCF event dispatcher has shut down
                pass

    def __traceMessageSent(self, m):
        for l in m.trace:
            try:
                tokenID = None
                if m.token is not None:
                    tokenID = m.token.getID()
                l.onMessageSent(m.type, tokenID, m.service, m.name, m.data)
            except Exception as x:
                x.tb = sys.exc_info()[2]
                protocol.log("Exception in channel listener", x)

    def start(self):
        assert protocol.isDispatchThread()
        protocol.invokeLater(self.__initServices)
        self.inp_thread.start()
        self.out_thread.start()

    def __initServices(self):
        try:
            if self.proxy:
                return
            if self.state == STATE_CLOSED:
                return
            services.onChannelCreated(self, self.local_service_by_name)
            self.__makeServiceByClassMap(self.local_service_by_name,
                                         self.local_service_by_class)
            args = list(self.local_service_by_name.keys())
            self.sendEvent(protocol.getLocator(), "Hello",
                           toJSONSequence((args,)))
        except IOError as x:
            self.terminate(x)

    def redirect_id(self, peer_id):
        """
        Redirect this channel to given peer using this channel remote peer
        locator service as a proxy.
        @param peer_id - peer that will become new remote communication
        endpoint of this channel.
        """
        peerMap = {}
        peerMap[peer.ATTR_ID] = peer_id
        self.redirect(map)

    def redirect(self, peer_attrs):
        """
        Redirect this channel to given peer using this channel remote peer
        locator service as a proxy.
        @param peer_attrs - peer that will become new remote communication
        endpoint of this channel.
        """
        if isinstance(peer_attrs, compat.strings):
            # support for redirect(peerId)
            attrs = {}
            attrs[peer.ATTR_ID] = peer_attrs
            peer_attrs = attrs
        channel = self
        assert protocol.isDispatchThread()
        if self.state == STATE_OPENING:
            self.redirect_queue.append(peer_attrs)
        else:
            assert self.state == STATE_OPEN
            assert self.redirect_command is None
            try:
                l = self.remote_service_by_class.get(locator.LocatorService)
                if not l:
                    raise IOError("Cannot redirect channel: peer " +
                                  self.remote_peer.getID() +
                                  " has no locator service")
                peer_id = peer_attrs.get(peer.ATTR_ID)
                if peer_id and len(peer_attrs) == 1:
                    _peer = l.getPeers().get(peer_id)
                    if not _peer:
                        # Peer not found, must wait for a while until peer is
                        # discovered or time out
                        class Callback(object):
                            found = None

                            def __call__(self):
                                if self.found:
                                    return
                                channel.terminate(Exception("Peer " + peer_id +
                                                            " not found"))
                        cb = Callback()
                        delay = locator.DATA_RETENTION_PERIOD / 3
                        protocol.invokeLaterWithDelay(delay, cb)

                        class Listener(locator.LocatorListener):
                            def peerAdded(self, new_peer):
                                if new_peer.getID() == peer_id:
                                    cb.found = True
                                    channel.state = STATE_OPEN
                                    l.removeListener(self)
                                    channel.redirect_id(peer_id)
                        l.addListener(Listener())
                    else:
                        class DoneRedirect(locator.DoneRedirect):
                            def doneRedirect(self, token, exc):
                                assert channel.redirect_command is token
                                channel.redirect_command = None
                                if channel.state != STATE_OPENING:
                                    return
                                if exc:
                                    channel.terminate(exc)
                                channel.remote_peer = _peer
                                channel.remote_service_by_class.clear()
                                channel.remote_service_by_name.clear()
                                channel.event_listeners.clear()
                        self.redirect_command = l.redirect(peer_id,
                                                           DoneRedirect())
                else:
                    class TransientPeer(peer.TransientPeer):
                        def __init__(self, peer_attrs, parent):
                            super(TransientPeer, self).__init__(peer_attrs)
                            self.parent = parent

                        def openChannel(self):
                            c = self.parent.openChannel()
                            c.redirect(peer_attrs)

                    class DoneRedirect(locator.DoneRedirect):
                        def doneRedirect(self, token, exc):
                            assert channel.redirect_command is token
                            channel.redirect_command = None
                            if channel.state != STATE_OPENING:
                                return
                            if exc:
                                channel.terminate(exc)
                            parent = channel.remote_peer
                            channel.remote_peer = TransientPeer(peer_attrs,
                                                                parent)
                            channel.remote_service_by_class.clear()
                            channel.remote_service_by_name.clear()
                            channel.event_listeners.clear()
                    self.redirect_command = l.redirect(peer_attrs,
                                                       DoneRedirect())
                self.state = STATE_OPENING
            except Exception as x:
                self.terminate(x)

    def __makeServiceByClassMap(self, by_name, by_class):
        for service in list(by_name.values()):
            by_class[service.__class__] = service
            for clazz in service.__class__.__bases__:
                if clazz == services.Service:
                    continue
                # TODO
                # if (!IService.class.isAssignableFrom(fs)) continue
                by_class[clazz] = service

    def getState(self):
        return self.state

    def addChannelListener(self, listener):
        assert protocol.isDispatchThread()
        assert listener
        self.channel_listeners.append(listener)

    def removeChannelListener(self, listener):
        assert protocol.isDispatchThread()
        self.channel_listeners.remove(listener)

    def addTraceListener(self, listener):
        if self.trace_listeners is None:
            self.trace_listeners = []
        else:
            self.trace_listeners = self.trace_listeners[:]
        self.trace_listeners.append(listener)

    def removeTraceListener(self, listener):
        self.trace_listeners = self.trace_listeners[:]
        self.trace_listeners.remove(listener)
        if len(self.trace_listeners) == 0:
            self.trace_listeners = None

    def addEventListener(self, service, listener):
        assert protocol.isDispatchThread()
        svc_name = str(service)
        listener.svc_name = svc_name
        lst = self.event_listeners.get(svc_name) or []
        lst.append(listener)
        self.event_listeners[svc_name] = lst

    def removeEventListener(self, service, listener):
        assert protocol.isDispatchThread()
        svc_name = str(service)
        lst = self.event_listeners.get(svc_name)
        if not lst:
            return
        for i in range(len(lst)):
            if lst[i] is listener:
                if len(lst) == 1:
                    del self.event_listeners[svc_name]
                else:
                    del lst[i]
                return

    def addCommandServer(self, service, listener):
        assert protocol.isDispatchThread()
        svc_name = str(service)
        if self.command_servers.get(svc_name):
            raise Exception("Only one command server per service is allowed")
        self.command_servers[svc_name] = listener

    def removeCommandServer(self, service, listener):
        assert protocol.isDispatchThread()
        svc_name = str(service)
        if self.command_servers.get(svc_name) is not listener:
            raise Exception("Invalid command server")
        del self.command_servers[svc_name]

    def close(self):
        assert protocol.isDispatchThread()
        if self.state == STATE_CLOSED:
            return
        try:
            for chan, listener in self.linked_channels.items():
                chan.removeChannelListener(listener)
            self.__sendEndOfStream(10000)
            self._close(None)
        except Exception as x:
            self._close(x)

    def terminate(self, error):
        assert protocol.isDispatchThread()
        if self.state == STATE_CLOSED:
            return
        try:
            self.__sendEndOfStream(500)
        except Exception as x:
            if not error:
                error = x
        self._close(error)

    def __sendEndOfStream(self, timeout):
        with self.out_lock:
            del self.out_queue[:]
            self.out_queue.append(None)
            self.out_lock.notify()
        self.out_thread.join(timeout)

    def _close(self, error):
        assert self.state != STATE_CLOSED
        self.state = STATE_CLOSED
        # Closing channel underlying streams can block for a long time,
        # so it needs to be done by a background thread.
        thread = threading.Thread(target=self.stop, name="TCF Channel Cleanup")
        thread.daemon = True
        thread.start()
        if error and isinstance(self.remote_peer, peer.AbstractPeer):
            self.remote_peer.onChannelTerminated()
        if self.registered_with_trasport:
            self.registered_with_trasport = False
            transport.channelClosed(self, error)
        if self.proxy:
            try:
                self.proxy.onChannelClosed(error)
            except Exception as x:
                x.tb = sys.exc_info()[2]
                protocol.log("Exception in channel listener", x)
        channel = self

        class Runnable(object):
            def __call__(self):
                if channel.out_tokens:
                    x = None
                    if isinstance(error, Exception):
                        x = error
                    elif error:
                        x = Exception(error)
                    else:
                        x = IOError("Channel is closed")
                    for msg in list(channel.out_tokens.values()):
                        try:
                            s = str(msg)
                            if len(s) > 72:
                                s = s[:72] + "...]"
                            y = IOError("Command " + s + " aborted")
#                            y.initCause(x)
                            msg.token.getListener().terminated(msg.token, y)
                        except Exception as e:
                            protocol.log("Exception in command listener", e)
                    channel.out_tokens.clear()
                if channel.channel_listeners:
                    for l in channel.channel_listeners:
                        if not l:
                            break
                        try:
                            l.onChannelClosed(error)
                        except Exception as x:
                            x.tb = sys.exc_info()[2]
                            protocol.log("Exception in channel listener", x)
                if channel.trace_listeners:
                    for l in channel.trace_listeners:
                        try:
                            l.onChannelClosed(error)
                        except Exception as x:
                            x.tb = sys.exc_info()[2]
                            protocol.log("Exception in channel listener", x)
                if error:
                    protocol.log("TCF channel terminated", error)
        protocol.invokeLater(Runnable())

    def getCongestion(self):
        assert protocol.isDispatchThread()
        level = len(self.out_tokens) * 100 / self.pending_command_limit - 100
        if self.remote_congestion_level > level:
            level = self.remote_congestion_level
        if level > 100:
            level = 100
        return level

    def getLocalPeer(self):
        assert protocol.isDispatchThread()
        return self.local_peer

    def getRemotePeer(self):
        assert protocol.isDispatchThread()
        return self.remote_peer

    def get_local_services(self):
        assert protocol.isDispatchThread()
        assert self.state != STATE_OPENING
        return list(self.local_service_by_name.keys())

    def getRemoteServices(self):
        assert protocol.isDispatchThread()
        assert self.state != STATE_OPENING
        return list(self.remote_service_by_name.keys())

    def get_local_service(self, cls_or_name):
        assert protocol.isDispatchThread()
        assert self.state != STATE_OPENING
        if isinstance(cls_or_name, compat.strings):
            return self.local_service_by_name.get(cls_or_name)
        else:
            return self.local_service_by_class.get(cls_or_name)

    def getRemoteService(self, cls_or_name):
        assert protocol.isDispatchThread()
        assert self.state != STATE_OPENING
        if isinstance(cls_or_name, compat.strings):
            return self.remote_service_by_name.get(cls_or_name)
        else:
            return self.remote_service_by_class.get(cls_or_name)

    def setServiceProxy(self, service_interface, service_proxy):
        if not self.notifying_channel_opened:
            raise Exception("setServiceProxe() can be called only from " +
                            "channel open call-back")
        proxy = self.remote_service_by_name.get(service_proxy.getName())
        if not isinstance(proxy, services.GenericProxy):
            raise Exception("Proxy already set")
        if self.remote_service_by_class.get(service_interface):
            raise Exception("Proxy already set")
        self.remote_service_by_class[service_interface] = service_proxy
        self.remote_service_by_name[service_proxy.getName()] = service_proxy

    def setProxy(self, proxy, services):
        self.proxy = proxy
        self.sendEvent(protocol.getLocator(), "Hello",
                       toJSONSequence((services,)))
        self.local_service_by_class.clear()
        self.local_service_by_name.clear()

    def addToOutQueue(self, msg):
        msg.trace = self.trace_listeners
        with self.out_lock:
            self.out_queue.append(msg)
            self.out_lock.notify()

    def sendCommand(self, service, name, args, listener):
        assert protocol.isDispatchThread()
        if self.state == STATE_OPENING:
            raise Exception("Channel is waiting for Hello message")
        if self.state == STATE_CLOSED:
            raise Exception("Channel is closed")
        msg = Message('C')
        msg.service = str(service)
        msg.name = name
        msg.data = args
        channel = self

        class CancelableToken(Token):
            def __init__(self, listener):
                super(CancelableToken, self).__init__(listener=listener)

            def cancel(self):
                assert protocol.isDispatchThread()
                if channel.state != STATE_OPEN:
                    return False
                with channel.out_lock:
                    if msg.is_sent:
                        return False
                    msg.is_canceled = True
                del channel.out_tokens[msg.token.getID()]
                return True
        token = CancelableToken(listener)
        msg.token = token
        self.out_tokens[token.getID()] = msg
        self.addToOutQueue(msg)
        return token

    def sendProgress(self, token, results):
        assert protocol.isDispatchThread()
        if self.state != STATE_OPEN:
            raise Exception("Channel is closed")
        msg = Message('P')
        msg.data = results
        msg.token = token
        self.addToOutQueue(msg)

    def sendResult(self, token, results):
        assert protocol.isDispatchThread()
        if self.state != STATE_OPEN:
            raise Exception("Channel is closed")
        msg = Message('R')
        msg.data = results
        msg.token = token
        self.addToOutQueue(msg)

    def rejectCommand(self, token):
        assert protocol.isDispatchThread()
        if self.state != STATE_OPEN:
            raise Exception("Channel is closed")
        msg = Message('N')
        msg.token = token
        self.addToOutQueue(msg)

    def sendEvent(self, service, name, args):
        assert protocol.isDispatchThread()
        if not (self.state == STATE_OPEN or self.state == STATE_OPENING and
                isinstance(service, locator.LocatorService)):
            raise Exception("Channel is closed")
        msg = Message('E')
        msg.service = str(service)
        msg.name = name
        msg.data = args
        self.addToOutQueue(msg)

    def isZeroCopySupported(self):
        return self.zero_copy

    def __traceMessageReceived(self, m):
        for l in self.trace_listeners:
            try:
                messageID = None
                if m.token is not None:
                    messageID = m.token.getID()
                l.onMessageReceived(m.type, messageID, m.service, m.name,
                                    m.data)
            except Exception as x:
                x.tb = sys.exc_info()[2]
                protocol.log("Exception in channel listener", x)

    def __handleInput(self, msg):
        assert protocol.isDispatchThread()
        if self.state == STATE_CLOSED:
            return
        if self.trace_listeners:
            self.__traceMessageReceived(msg)
        try:
            token = None
            typeCode = msg.type
            if typeCode in 'PRN':
                token_id = msg.token.getID()
                cmd = self.out_tokens.get(token_id)
                if cmd is None:
                    raise Exception("Invalid token received: " + token_id)
                if typeCode != 'P':
                    del self.out_tokens[token_id]
                token = cmd.token
            if typeCode == 'C':
                if self.state == STATE_OPENING:
                    raise IOError("Received command " + msg.service + "." +
                                  msg.name + " before Hello message")
                if self.proxy:
                    self.proxy.onCommand(msg.token, msg.service, msg.name,
                                         msg.data)
                else:
                    token = msg.token
                    cmds = self.command_servers.get(msg.service)
                    if cmds:
                        cmds.command(token, msg.name, msg.data)
                    else:
                        self.rejectCommand(token)
            elif typeCode == 'P':
                token.getListener().progress(token, msg.data)
                self.__sendCongestionLevel()
            elif typeCode == 'R':
                token.getListener().result(token, msg.data)
                self.__sendCongestionLevel()
            elif typeCode == 'N':
                report = errors.ErrorReport("Command is not recognized",
                                            errors.TCF_ERROR_INV_COMMAND)
                token.getListener().terminated(token, report)
            elif typeCode == 'E':
                hello = msg.service == locator.NAME and msg.name == "Hello"
                if hello:
                    self.remote_service_by_name.clear()
                    self.remote_service_by_class.clear()
                    data = fromJSONSequence(msg.data)[0]
                    services.onChannelOpened(self, data,
                                             self.remote_service_by_name)
                    self.__makeServiceByClassMap(self.remote_service_by_name,
                                                 self.remote_service_by_class)
                    self.zero_copy = "ZeroCopy" in self.remote_service_by_name
                if self.proxy and self.state == STATE_OPEN:
                    self.proxy.onEvent(msg.service, msg.name, msg.data)
                elif hello:
                    assert self.state == STATE_OPENING
                    self.state = STATE_OPEN
                    assert self.redirect_command is None
                    if self.redirect_queue:
                        self.redirect(self.redirect_queue.pop(0))
                    else:
                        self.notifying_channel_opened = True
                        if not self.registered_with_trasport:
                            transport.channelOpened(self)
                            self.registered_with_trasport = True
                        for l in tuple(self.channel_listeners):
                            if not l:
                                break
                            try:
                                l.onChannelOpened()
                            except Exception as x:
                                x.tb = sys.exc_info()[2]
                                protocol.log("Exception in channel listener",
                                             x)
                        self.notifying_channel_opened = False
                else:
                    lst = self.event_listeners.get(msg.service)
                    if lst:
                        for l in lst:
                            l.event(msg.name, msg.data)
                    self.__sendCongestionLevel()
            elif typeCode == 'F':
                length = len(msg.data)
                if length > 0 and msg.data[length - 1] == 0:
                    length -= 1
                self.remote_congestion_level = int(msg.data[:length])
            else:
                assert False
        except Exception as x:
            x.tb = sys.exc_info()[2]
            self.terminate(x)

    def __sendCongestionLevel(self):
        self.local_congestion_cnt += 1
        if self.local_congestion_cnt < 8:
            return
        self.local_congestion_cnt = 0
        if self.state != STATE_OPEN:
            return
        timeVal = int(time.time() * 1000)
        if timeVal - self.local_congestion_time < 500:
            return
        assert protocol.isDispatchThread()
        level = protocol.getCongestionLevel()
        if level == self.local_congestion_level:
            return
        i = int((level - self.local_congestion_level) / 8)
        if i != 0:
            level = self.local_congestion_level + i
        self.local_congestion_time = timeVal
        with self.out_lock:
            msg = None
            if self.out_queue:
                msg = self.out_queue[0]
            if msg is None or msg.type != 'F':
                msg = Message('F')
                self.out_queue.insert(0, msg)
                self.out_lock.notify()
            data = "%i\0" % self.local_congestion_level
            msg.data = data
            msg.trace = self.trace_listeners
            self.local_congestion_level = level

    def read(self):
        """
        Read one byte from the channel input stream.
        @return next data byte or EOS (-1) if end of stream is reached,
        or EOM (-2) if end of message is reached.
        @raises IOError
        """
        raise NotImplementedError("Abstract method")

    def writeByte(self, n):
        """
        Write one byte into the channel output stream.
        The method argument can be one of two special values:
        - EOS (-1) end of stream marker
        - EOM (-2) end of message marker.
        The stream can put the byte into a buffer instead of transmitting it
        right away.
        @param n - the data byte.
        @raises IOError
        """
        raise NotImplementedError("Abstract method")

    def flush(self):
        """
        Flush the channel output stream.
        All buffered data should be transmitted immediately.
        @raises IOError
        """
        raise NotImplementedError("Abstract method")

    def stop(self):
        """
        Stop (close) channel underlying streams.
        If a thread is blocked by read() or write(), it should be
        resumed (or interrupted).
        @raises IOError
        """
        raise NotImplementedError("Abstract method")

    def write(self, buf):
        """
        Write array of bytes into the channel output stream.
        The stream can put bytes into a buffer instead of transmitting it right
        away.
        @param buf
        @raises IOError
        """
        assert threading.currentThread() == self.out_thread
        for i in buf:
            self.writeByte(ord(buf[i]) & 0xff)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from chipscopy import CoreType
from chipscopy.api._detail.debug_core import DebugCore
from chipscopy.client.axis_trace_core_client import AxisTraceClient


class Trace(DebugCore["AxisTraceClient"]):
    def __init__(self, tcf_trace: AxisTraceClient):
        DebugCore.__init__(self, core_type=CoreType.AXIS_TRACE, core_tcf_node=tcf_trace)
        self.filter_by = {"uuid": self.core_info.uuid}



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from typing import Optional, Dict, Any, List, Union, Callable

from chipscopy.dm import chipscope
from chipscopy.utils.logger import log
from chipscopy.utils.version import version_consistency_check
from chipscopy.client import connect, disconnect
from chipscopy.client.util import connect_hw
from chipscopy.client.view_info import TargetFilter, ViewInfo
from chipscopy.client.server_info import ServerInfo
from chipscopy.api.containers import QueryList
from chipscopy.api.device.device import Device
from chipscopy.api.device.device_scanner import DeviceScanner
from chipscopy.api.memory import Memory

DOMAIN_NAME = "client"

# Global session tracking here. We keep a list of who is connected at any time.
# Format:
#    dict[hex(id(session))] = (hw_server, cs_server, session)
_connected_session_dict: dict = {}


class Session:
    """Top level object that tracks a connection to a hardware server and optionally, chipscope
    server. To create and destroy a session, use the factory function
    create_session() and delete_session().
    """

    _hw_server_url: str = None
    _cs_server_url: str = None
    _xvc_server_url: str = None
    _hw_server: Optional[ServerInfo] = None
    _cs_server: Optional[ServerInfo] = None

    def __init__(
        self,
        *,
        hw_server_url: str,
        cs_server_url: Optional[str] = None,
        xvc_server_url: Optional[str] = None,
        disable_core_scan: bool,
        bypass_version_check: bool,
    ):
        self._disable_core_scan = disable_core_scan
        self._bypass_version_check = bypass_version_check
        self._hw_server_url = hw_server_url
        self._cs_server_url = cs_server_url
        self._xvc_server_url = xvc_server_url
        self.connect()

    def __str__(self):
        return f"{self.handle}"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        delete_session(self)

    @classmethod
    def _add_connection(
        cls, hw_server: ServerInfo, cs_server: Optional[ServerInfo], session: "Session"
    ):
        _connected_session_dict[hex(id(session))] = (hw_server, cs_server, session)

    @classmethod
    def _remove_connection(cls, session: "Session"):
        del _connected_session_dict[hex(id(session))]

    @classmethod
    def disconnect_all_sessions(cls):
        sessions_to_disconnect = []
        for id_key, (_, _, session) in _connected_session_dict.items():
            sessions_to_disconnect.append(session)
        for session in sessions_to_disconnect:
            try:
                session.disconnect()
            except Exception:  # pragma: no cover
                # TODO: Look into root cause later - happens intermittently in pytest infra causing tests to fail
                pass

    def connect(self):
        if not self._hw_server_url:
            raise ValueError("hw_server_url must point to a valid hw_server")
        self._hw_server = Session._connect_hw_server(self._hw_server_url)
        if self._cs_server_url:
            self._cs_server = Session._connect_cs_server(self._cs_server_url)
            self._cs_server.connect_remote(self._hw_server.url)
            if self._xvc_server_url:
                self._cs_server.connect_xvc(self._xvc_server_url, self._hw_server_url)
        Session._add_connection(self._hw_server, self._cs_server, self)
        try:
            # Quick sanity check - throws RuntimeError on version mismatch
            self._version_check()
        except RuntimeError:
            self.disconnect()
            t, v, tb = sys.exc_info()
            raise t(v).with_traceback(tb)

    def disconnect(self):
        Session._remove_connection(self)
        if self._cs_server:
            self._cs_server.disconnect_remote(f"TCP:{self._hw_server.url}")
            disconnect(self._cs_server)
            self._cs_server = None
        disconnect(self._hw_server)
        self._hw_server = None

    def set_param(self, params: Dict[str, Any]):
        """Generic parameter get and set for low level chipscope server params"""
        if not isinstance(params, dict):
            message = "Please provide the params to set as a dictionary!"
            log[DOMAIN_NAME].error(message)
            raise TypeError(message)
        cs_service = self._cs_server.get_sync_service("ChipScope")
        cs_service.set_css_param(params)

    def get_param(self, params: Union[str, List[str]]) -> Dict[str, str]:
        """Generic parameter get and set for low level chipscope server params"""
        if isinstance(params, str):
            params = [params]
        cs_service = self._cs_server.get_sync_service("ChipScope")
        return cs_service.get_css_param(params).get()

    @property
    def hw_server(self) -> ServerInfo:
        return self._hw_server

    @property
    def cs_server(self) -> ServerInfo:
        return self._cs_server

    @property
    def handle(self) -> str:
        if self._cs_server and self._hw_server:
            handle_str = f"{self._cs_server.url}<->{self._hw_server.url}"
        elif self._hw_server:
            handle_str = f"{self._hw_server.url}"
        else:
            handle_str = "no_hw_server<->no_cs_server"
        return handle_str

    @property
    def chipscope_view(self) -> ViewInfo:
        view = None
        if self._cs_server:
            view = self._cs_server.get_view(chipscope)
        return view

    @property
    def devices(self) -> QueryList[Device]:
        """Returns a list of devices connected to this hw_server. Devices may
        contain several chains so don't always assume a single jtag chain
        in-order is returned"""
        devices: QueryList[Device] = QueryList()
        device_scanner = DeviceScanner(self._hw_server, self._cs_server)
        device_scanner.scan_devices()
        for device_identification in device_scanner.get_scan_results():
            device_node = device_identification.find_top_level_device_node(self._hw_server)
            if device_node.device_wrapper is None:
                device_node.device_wrapper = Device(
                    hw_server=self._hw_server,
                    cs_server=self._cs_server,
                    device_identification=device_identification,
                    disable_core_scan=self._disable_core_scan,
                )
            devices.append(device_node.device_wrapper)
        return devices

    @property
    def memory(self) -> QueryList[Memory]:
        memory_node_list = QueryList()
        # This is not fast - improve it later. We don't need to query each
        # device - but it's clear and simple now.
        for device in self.devices:
            memory_nodes = device.memory
            for node in memory_nodes:
                memory_node_list.append(node)
        return memory_node_list

    @staticmethod
    def set_log_level(level: str = None):
        """
        Set the logging level for the ChipScoPy client. This applies to all sessions.
        Default is "NONE"

        Args:
            level: The minimum level to use for the logger. Valid levels are
                "TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL", "NONE"

        """

        if level is None:
            log.disable_domain(DOMAIN_NAME)

        valid_levels = ["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL", "NONE"]
        level = level.upper()
        assert level in valid_levels
        log.change_log_level(level)
        log.enable_domain(DOMAIN_NAME)

    @staticmethod
    def _connect_hw_server(hw_server_url: str) -> ServerInfo:
        return Session._connect_server("hw_server", hw_server_url, connect_hw)

    @staticmethod
    def _connect_cs_server(cs_server_url: str) -> ServerInfo:
        return Session._connect_server("cs_server", cs_server_url, connect)

    @staticmethod
    def _connect_server(server_name: str, server_url: str, connect_func: Callable) -> ServerInfo:
        connection_refused_msg = None
        try:
            server: ServerInfo = connect_func(server_url)
        except ConnectionRefusedError:
            # Shorten stack dump, by creating new exception.
            connection_refused_msg = (
                f"Connection could not be opened to {server_name} @ {server_url}.\n"
                f"  Please make sure that {server_name} is running at the URL provided, and firewall is not blocking."
            )
        except Exception as ex:
            # Chain exceptions.
            raise Exception(
                f"Connection could not be opened to {server_name} @ {server_url}."
            ) from ex

        if connection_refused_msg:
            raise ConnectionRefusedError(connection_refused_msg)
        return server

    def _version_check(self):
        version_consistency_check(self._hw_server, self._cs_server, self._bypass_version_check)


###############################################################################
# Factory methods below
###############################################################################


def create_session(*, hw_server_url: str, cs_server_url: Optional[str] = None, **kwargs) -> Session:
    """
    Create a new session. Connect to the specified hw_server, and optionally
    the cs_server.

    - hw_server is used for programming and Versal Memory read/write operations
    - cs_server is used for higher level debug core communication

    Example 1: Default session create ::

        my_session = create_session(hw_server_url="TCP:localhost:3121",
                                    cs_server_url="TCP:localhost:3042")

    Example 2: Disable core scanning and server version checking ::

        my_session = create_session(hw_server_url="TCP:localhost:3121",
                                    cs_server_url="TCP:localhost:3042",
                                    disable_core_scan=True,
                                    bypass_version_check=True)

    Args:
        hw_server_url: Hardware server URL. Format ``TCP:<hostname>:<port>``
        cs_server_url: ChipScope server URL. Format ``TCP:<hostname>:<port>``

    Returns:
        New session object.

    """
    # Optional arguments in kwargs -
    #    disable_core_scan - This determines how devices setup the debug cores.
    #         It is possible to disable the setup_debug_cores detection
    #         if it is likely to interfere with an existing hw_server
    disable_core_scan = kwargs.get("disable_core_scan", False)
    bypass_version_check = kwargs.get("bypass_version_check", False)
    xvc_server_url = kwargs.get("xvc_server_url", None)

    # Create session even if there already exists a session with the same cs_server and hw_server
    # It *should* be safe.
    return Session(
        cs_server_url=cs_server_url,
        hw_server_url=hw_server_url,
        xvc_server_url=xvc_server_url,
        disable_core_scan=disable_core_scan,
        bypass_version_check=bypass_version_check,
    )


def delete_session(session: Session):
    """
    Delete a session. Shuts down any open server connections.

    Args:
        session: The session object to delete
    """
    session.disconnect()



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from dataclasses import dataclass, field
from typing import Callable, Dict, Any, Optional, TYPE_CHECKING, Set, List

from chipscopy.api.ibert.aliases import (
    MB_ELF_VERSION,
    YK_SCAN_START_TIME,
    YK_SCAN_STOP_TIME,
    YK_SCAN_SLICER_DATA,
    YK_SCAN_SNR_VALUE,
)
from chipscopy.api.ibert.rx import RX
from chipscopy.utils.printer import printer

if TYPE_CHECKING:
    from chipscopy.dm import Node


@dataclass
class YKSample:
    slicer: List[float]
    snr: float


@dataclass
class YKScan:
    """
    Class for interacting with YK scans.
    **Please do not** create an instance of this class directly. Please use the factory method
    :py:func:`~chipscopy.api.ibert.create_yk_scans` instead.
    """

    rx: RX
    """:py:class:`RX` object attached to this eye scan"""

    name: str
    """Name of the eye scan"""

    updates_callback: Callable[["YKScan"], None] = None
    """Callback function called when eye scan has ended"""

    filter_by: Dict[str, Any] = field(default_factory=dict)

    scan_data: List[YKSample] = field(default_factory=list)
    """YK scan data samples in the order they are received"""

    stop_time: datetime = None
    """Time stamp of when eye scan was stopped in cs_server"""

    start_time: datetime = None
    """Time stamp of when eye scan was started in cs_server"""

    elf_version: str = None
    """ELF version read from the MicroBlaze"""

    _handle_from_cs_server: Optional[str] = None

    def __repr__(self):
        return self.name

    def __post_init__(self):
        self.rx.yk_scan = self

        self.filter_by = {"rx": self.rx, "name": self.name}

        self.rx.property.endpoint_tcf_node.add_listener(self._update_event_listener)

    def start(self):
        self._handle_from_cs_server = self.rx.core_tcf_node.start_yk_scan(rx_name=self.rx.handle)

    def _update_event_listener(self, node: "Node", updated_properties: Set[str]):
        # NOTE - This is called on the TCF event dispatcher thread
        if len(updated_properties) == 0 or self._handle_from_cs_server not in updated_properties:
            return

        try:
            report = node.props[self._handle_from_cs_server]

            if self.start_time is None and YK_SCAN_START_TIME in report:
                self.start_time = datetime.strptime(
                    report[YK_SCAN_START_TIME], "%Y-%m-%d %H:%M:%S.%f"
                )

            # Not expected to change after scan stop
            if self.stop_time is None and YK_SCAN_STOP_TIME in report:
                self.stop_time = datetime.strptime(
                    report[YK_SCAN_STOP_TIME], "%Y-%m-%d %H:%M:%S.%f"
                )

            # Not expected to change after scan start
            if self.elf_version is None and MB_ELF_VERSION in report:
                self.elf_version = report[MB_ELF_VERSION]

            if YK_SCAN_SLICER_DATA in report and YK_SCAN_SNR_VALUE in report:
                self.scan_data.append(
                    YKSample(report[YK_SCAN_SLICER_DATA], report[YK_SCAN_SNR_VALUE])
                )

            # If user has registered done callback function call it.
            if callable(self.updates_callback):
                try:
                    self.updates_callback(self)
                except Exception as e:
                    printer(
                        f"Unhandled exception during YK scan update callback!\n"
                        f"Exception - {str(e)}",
                        level="warning",
                    )

        except Exception as e:
            printer(
                f"Unhandled YK scan update exception on TCF thread!\nException - {str(e)}",
                level="warning",
            )

    def stop(self):
        """
        Stop eye scan, that is in-progress in the MicroBlaze

        """
        self.rx.core_tcf_node.terminate_yk_scan(rx_name=self.rx.handle)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import List, Optional, Union


@dataclass
class EyeScanParam:
    name: str
    """Name of the eye scan param"""

    modifiable: bool = False
    """Can the parameter value be modified"""

    valid_values: List[Union[int, str]] = None
    """Valid values for the parameter, if it can be modified"""

    default_value: Union[int, str] = None
    """Default value of the parameter"""

    value: Optional[Union[int, str]] = None
    """Value set by the user. Taken into consideration only if the param is modifiable"""



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from chipscopy import dm
from chipscopy.dm import request
from chipscopy.dm.chipscope import ChipScopeManager
from chipscopy.tcf import protocol


class ExampleMemNode(dm.Node):
    @staticmethod
    def is_compatible(node: "Node"):
        if not node.manager:
            return node.type
        if not node.parent:
            return True
        parent = node.manager[node.parent]
        if parent.type and parent.type == "debug_hub":  # does not work for cores under a DebugHub
            return False
        return isinstance(node.manager, ChipScopeManager)

    def post_init(self):
        self["mem_setup"] = True
        if self.dud_was_here:
            self["dud_was_here"] = False

        self.mrd(0x100, done=self.default_done())  # implicitly initializes node in cs_server

    def default_done(self, done=None):
        def done_cmd(token, error, results):
            self.remove_pending(token)
            if done:
                done(token, error, results)

        return done_cmd

    def get_example_service(self):
        if not self.manager:
            raise Exception("Unable to run memory operation")

        service = self.manager.channel.getRemoteService("Example")

        if not service:
            raise Exception("Example Service not supported in server")

        return service

    def mrd(self, addr: int, word_count: int = 1, done: request.DoneCallback = None):
        service = self.get_example_service()
        return self.add_pending(service.mrd(self.ctx, addr, word_count, self.default_done(done)))

    def mwr(
        self, addr: int, words: List[int], word_count: int = 0, done: request.DoneCallback = None
    ):
        service = self.get_example_service()
        return self.add_pending(
            service.mwr(self.ctx, addr, words, word_count, self.default_done(done))
        )


class ExampleDudNode(dm.Node):
    def __init__(self):
        super().__init__()
        self.duded = 0

    @staticmethod
    def is_compatible(node: "Node"):
        return (not node.manager) or isinstance(node.manager, ChipScopeManager)

    def post_init(self):
        self["dud_was_here"] = True
        self.duded += 1
        self["duded"] = self.duded

    @staticmethod
    def default_done(done=None):
        def done_cmd(token, error, results):
            if done:
                done(token, error, results)

        return protocol.invokeLater(done_cmd)

    def mrd(self, addr: int, word_count: int = 1, done: request.DoneCallback = None):
        return self.default_done(done)

    def mwr(
        self, addr: int, words: List[int], word_count: int = 0, done: request.DoneCallback = None
    ):
        return self.default_done(done)
